from dataclasses import dataclass
from enum import Enum
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.plate_reader.rec._2024._06.plate_reader import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocument,
    Model,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
    TQuantityValueModel,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValuePicogramPerMilliliter,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueRelativeLightUnit,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.models.shared.definitions.definitions import TDatacube
from allotropy.allotrope.schema_mappers.data_cube import DataCube, get_data_cube
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.utils.values import (
    assert_not_none,
    quantity_or_none,
)


# TODO: these two enum classes should come from the schema
class ScanPositionSettingPlateReader(Enum):
    bottom_scan_position__plate_reader_ = "bottom scan position (plate reader)"
    scan_position_configuration__plate_reader_ = (
        "scan position configuration (plate reader)"
    )
    top_scan_position__plate_reader_ = "top scan position (plate reader)"


class ContainerType(Enum):
    reactor = "reactor"
    controlled_lab_reactor = "controlled lab reactor"
    tube = "tube"
    well_plate = "well plate"
    differential_scanning_calorimetry_pan = "differential scanning calorimetry pan"
    qPCR_reaction_block = "qPCR reaction block"  # noqa: N815
    vial_rack = "vial rack"
    pan = "pan"
    reservoir = "reservoir"
    array_card_block = "array card block"
    capillary = "capillary"
    disintegration_apparatus_basket = "disintegration apparatus basket"
    jar = "jar"
    container = "container"
    tray = "tray"
    basket = "basket"
    cell_holder = "cell holder"


class MeasurementType(Enum):
    ULTRAVIOLET_ABSORBANCE = "ULTRAVIOLET_ABSORBANCE"
    FLUORESCENCE = "FLUORESCENCE"
    LUMINESCENCE = "LUMINESCENCE"
    ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR = "ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR"
    FLUORESCENCE_CUBE_DETECTOR = "FLUORESCENCE_CUBE_DETECTOR"
    LUMINESCENCE_CUBE_DETECTOR = "LUMINESCENCE_CUBE_DETECTOR"


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]
    description: str | None = None


@dataclass(frozen=True)
class ErrorDocument:
    error: str
    error_feature: str


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    type_: MeasurementType
    device_type: str
    identifier: str
    sample_identifier: str
    location_identifier: str
    firmware_version: str | None = None
    batch_identifier: str | None = None

    # Optional metadata
    well_plate_identifier: str | None = None
    well_location_identifier: str | None = None
    detection_type: str | None = None
    sample_role_type: SampleRoleType | None = None
    mass_concentration: float | None = None

    # Measurements
    absorbance: float | None = None
    fluorescence: float | None = None
    luminescence: float | None = None

    # Settings
    detector_wavelength_setting: float | None = None
    detector_bandwidth_setting: float | None = None
    excitation_wavelength_setting: float | None = None
    excitation_bandwidth_setting: float | None = None
    wavelength_filter_cutoff_setting: float | None = None
    detector_distance_setting: float | None = None
    scan_position_setting: ScanPositionSettingPlateReader | None = None
    detector_gain_setting: str | None = None
    detector_carriage_speed: str | None = None
    compartment_temperature: float | None = None
    number_of_averages: float | None = None
    measurement_custom_info: dict[str, Any] | None = None
    sample_custom_info: dict[str, Any] | None = None
    device_control_custom_info: dict[str, Any] | None = None
    electronic_absorbance_reference_wavelength_setting: float | None = None

    # Kinetic settings
    total_measurement_time_setting: float | None = None
    read_interval_setting: float | None = None
    number_of_scans_setting: float | None = None

    # Kinetic measurements
    profile_data_cube: DataCube | None = None

    # error documents
    error_document: list[ErrorDocument] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: float
    measurement_time: str
    analyst: str | None = None
    analytical_method_identifier: str | None = None
    experimental_data_identifier: str | None = None
    experiment_type: str | None = None


@dataclass(frozen=True)
class Metadata:
    asm_file_identifier: str
    device_identifier: str
    model_number: str
    data_system_instance_id: str
    software_name: str | None = None
    unc_path: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None
    file_name: str | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/plate-reader/REC/2024/06/plate-reader.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.metadata.device_identifier,
                    model_number=data.metadata.model_number,
                    equipment_serial_number=data.metadata.equipment_serial_number,
                    product_manufacturer=data.metadata.product_manufacturer,
                ),
                data_system_document=DataSystemDocument(
                    ASM_file_identifier=data.metadata.asm_file_identifier,
                    data_system_instance_identifier=data.metadata.data_system_instance_id,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                plate_reader_document=[
                    self._get_technique_document(measurement_group)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data.calculated_data
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup
    ) -> PlateReaderDocumentItem:
        return PlateReaderDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                analytical_method_identifier=measurement_group.analytical_method_identifier,
                experimental_data_identifier=measurement_group.experimental_data_identifier,
                experiment_type=measurement_group.experiment_type,
                container_type=ContainerType.well_plate.value,
                plate_well_count=TQuantityValueNumber(
                    value=measurement_group.plate_well_count
                ),
                measurement_time=self.get_date_time(measurement_group.measurement_time),
                measurement_document=[
                    self._get_measurement_document(measurement)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        # TODO(switch-statement): use switch statement once Benchling can use 3.10 syntax
        if measurement.type_ == MeasurementType.ULTRAVIOLET_ABSORBANCE:
            return self._get_ultraviolet_absorbance_measurement_document(measurement)
        elif measurement.type_ == MeasurementType.LUMINESCENCE:
            return self._get_luminescence_measurement_document(measurement)
        elif measurement.type_ == MeasurementType.FLUORESCENCE:
            return self._get_fluorescence_measurement_document(measurement)
        elif measurement.type_ in [
            MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR,
            MeasurementType.LUMINESCENCE_CUBE_DETECTOR,
            MeasurementType.FLUORESCENCE_CUBE_DETECTOR,
        ]:
            return self._get_profile_data_cube_measurement_document(measurement)
        else:
            msg = f"Unexpected measurement type: {measurement.type_}"
            raise AllotropyParserError(msg)

    def _get_ultraviolet_absorbance_measurement_document(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        device_control_document = DeviceControlDocumentItem(
            electronic_absorbance_reference_wavelength_setting=quantity_or_none(
                TQuantityValueNanometer,
                measurement.electronic_absorbance_reference_wavelength_setting,
            ),
            device_type=measurement.device_type,
            detection_type=measurement.detection_type,
            detector_wavelength_setting=quantity_or_none(
                TQuantityValueNanometer,
                measurement.detector_wavelength_setting,
            ),
            number_of_averages=quantity_or_none(
                TQuantityValueNumber, measurement.number_of_averages
            ),
            detector_carriage_speed_setting=measurement.detector_carriage_speed,
            detector_gain_setting=measurement.detector_gain_setting,
            detector_distance_setting__plate_reader_=quantity_or_none(
                TQuantityValueMillimeter,
                measurement.detector_distance_setting,
            ),
            firmware_version=measurement.firmware_version,
        )
        measurement_doc = MeasurementDocument(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    add_custom_information_document(
                        device_control_document, measurement.device_control_custom_info
                    )
                ]
            ),
            absorbance=TQuantityValueMilliAbsorbanceUnit(
                value=assert_not_none(
                    value=measurement.absorbance,
                    msg="Missing absorbance value in ultraviolet absorbance measurement",
                )
            ),
            compartment_temperature=quantity_or_none(
                TQuantityValueDegreeCelsius, measurement.compartment_temperature
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.error_document
            ),
        )
        return add_custom_information_document(
            measurement_doc, measurement.measurement_custom_info
        )

    def _get_luminescence_measurement_document(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        doc = MeasurementDocument(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=measurement.device_type,
                        detection_type=measurement.detection_type,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_wavelength_setting,
                        ),
                        detector_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_bandwidth_setting,
                        ),
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter,
                            measurement.detector_distance_setting,
                        ),
                        scan_position_setting__plate_reader_=(
                            measurement.scan_position_setting.value
                            if measurement.scan_position_setting
                            else None
                        ),
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, measurement.number_of_averages
                        ),
                        detector_gain_setting=measurement.detector_gain_setting,
                        detector_carriage_speed_setting=measurement.detector_carriage_speed,
                    )
                ]
            ),
            luminescence=TQuantityValueRelativeLightUnit(
                value=assert_not_none(
                    measurement.luminescence,
                    msg="Missing luminescence value in luminescence measurement",
                )
            ),
            compartment_temperature=quantity_or_none(
                TQuantityValueDegreeCelsius, measurement.compartment_temperature
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.error_document
            ),
        )
        return add_custom_information_document(doc, measurement.measurement_custom_info)

    def _get_fluorescence_measurement_document(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        return MeasurementDocument(
            measurement_identifier=measurement.identifier,
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=measurement.device_type,
                        detection_type=measurement.detection_type,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_wavelength_setting,
                        ),
                        detector_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_bandwidth_setting,
                        ),
                        excitation_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.excitation_wavelength_setting,
                        ),
                        excitation_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.excitation_bandwidth_setting,
                        ),
                        wavelength_filter_cutoff_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.wavelength_filter_cutoff_setting,
                        ),
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter,
                            measurement.detector_distance_setting,
                        ),
                        scan_position_setting__plate_reader_=(
                            measurement.scan_position_setting.value
                            if measurement.scan_position_setting
                            else None
                        ),
                        detector_gain_setting=measurement.detector_gain_setting,
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, measurement.number_of_averages
                        ),
                        detector_carriage_speed_setting=measurement.detector_carriage_speed,
                    )
                ]
            ),
            fluorescence=TQuantityValueRelativeFluorescenceUnit(
                value=assert_not_none(
                    measurement.fluorescence,
                    msg="Missing fluorescence value in fluorescence measurement",
                )
            ),
            compartment_temperature=quantity_or_none(
                TQuantityValueDegreeCelsius, measurement.compartment_temperature
            ),
            sample_document=self._get_sample_document(measurement),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.error_document
            ),
        )

    def _get_profile_data_cube_measurement_document(
        self, measurement: Measurement
    ) -> MeasurementDocument:
        if not measurement.profile_data_cube:
            msg = "Profile data cube is required for cube detector measurements"
            raise AllotropyParserError(msg)

        profile_data_cube = get_data_cube(measurement.profile_data_cube, TDatacube)

        return MeasurementDocument(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=measurement.device_type,
                        detection_type=measurement.detection_type,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_wavelength_setting,
                        ),
                        excitation_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.excitation_wavelength_setting,
                        ),
                        wavelength_filter_cutoff_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.wavelength_filter_cutoff_setting,
                        ),
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, measurement.number_of_averages
                        ),
                        detector_carriage_speed_setting=measurement.detector_carriage_speed,
                        detector_gain_setting=measurement.detector_gain_setting,
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter,
                            measurement.detector_distance_setting,
                        ),
                        total_measurement_time_setting=quantity_or_none(
                            TQuantityValueSecondTime,
                            measurement.total_measurement_time_setting,
                        ),
                        read_interval_setting=quantity_or_none(
                            TQuantityValueSecondTime,
                            measurement.read_interval_setting,
                        ),
                        number_of_scans_setting=quantity_or_none(
                            TQuantityValueNumber,
                            measurement.number_of_scans_setting,
                        ),
                    )
                ]
            ),
            compartment_temperature=quantity_or_none(
                TQuantityValueDegreeCelsius, measurement.compartment_temperature
            ),
            absorption_profile_data_cube=(
                profile_data_cube
                if measurement.type_
                == MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR
                else None
            ),
            luminescence_profile_data_cube=(
                profile_data_cube
                if measurement.type_ == MeasurementType.LUMINESCENCE_CUBE_DETECTOR
                else None
            ),
            fluorescence_emission_profile_data_cube=(
                profile_data_cube
                if measurement.type_ == MeasurementType.FLUORESCENCE_CUBE_DETECTOR
                else None
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        sample_doc = SampleDocument(
            sample_identifier=measurement.sample_identifier,
            location_identifier=measurement.location_identifier,
            well_plate_identifier=measurement.well_plate_identifier,
            well_location_identifier=measurement.well_location_identifier,
            sample_role_type=(
                measurement.sample_role_type.value
                if measurement.sample_role_type
                else None
            ),
            mass_concentration=quantity_or_none(
                TQuantityValuePicogramPerMilliliter,
                measurement.mass_concentration,
            ),
            batch_identifier=measurement.batch_identifier,
        )
        return add_custom_information_document(
            sample_doc, measurement.sample_custom_info
        )

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDataItem] | None
    ) -> CalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    calculation_description=calculated_data_item.description,
                    calculated_result=TQuantityValueModel(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )

    def _get_error_aggregate_document(
        self, error_documents: list[ErrorDocument] | None
    ) -> ErrorAggregateDocument | None:
        return (
            ErrorAggregateDocument(
                error_document=[
                    ErrorDocumentItem(
                        error=error.error,
                        error_feature=error.error_feature,
                    )
                    for error in error_documents
                ]
            )
            if error_documents
            else None
        )



from __future__ import annotations

from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.chemometec_nucleoview.constants import (
    DEFAULT_ANALYST,
    DEFAULT_MODEL_NUMBER,
    NUCLEOCOUNTER_DETECTION_TYPE,
    NUCLEOCOUNTER_DEVICE_TYPE,
    NUCLEOCOUNTER_SOFTWARE_NAME,
)
from allotropy.parsers.constants import DEFAULT_EPOCH_TIMESTAMP, NOT_APPLICABLE
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(data: SeriesData, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=data.get(str, "PC", NOT_APPLICABLE),
        file_name=Path(file_path).name,
        unc_path=file_path,
        model_number=data.get(str, "Instrument type", DEFAULT_MODEL_NUMBER),
        equipment_serial_number=data.get(str, "Instrument s/n"),
        software_name=NUCLEOCOUNTER_SOFTWARE_NAME,
        software_version=data.get(str, "Application SW version"),
        device_type=NUCLEOCOUNTER_DEVICE_TYPE,
        detection_type=NUCLEOCOUNTER_DETECTION_TYPE,
    )


def create_measurement_groups(data: SeriesData) -> MeasurementGroup:
    timestamp = data.get(str, "Date time")
    if timestamp:
        offset = data.get(str, "Time zone offset", "0")
        timestamp = pd.to_datetime(
            f"{timestamp}{'+' if offset[0] not in {'+', '-'} else ''}{offset}"
        ).isoformat()

    def _converted_value_or_none(key: str) -> float | None:
        return value / 1e6 if (value := data.get(float, key)) is not None else None

    return MeasurementGroup(
        analyst=data.get(str, "Operator", DEFAULT_ANALYST),
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=timestamp or DEFAULT_EPOCH_TIMESTAMP,
                sample_identifier=data[str, "Image"].split("-")[3],
                cell_density_dilution_factor=data.get(float, "Multiplication factor"),
                viability=data[float, "Viability (%)"],
                # Cell counts are measured in cells/mL, but reported in millions of cells/mL
                viable_cell_density=data[float, "Live (cells/ml)"] / 1e6,
                dead_cell_density=_converted_value_or_none("Dead (cells/ml)"),
                total_cell_density=_converted_value_or_none("Total (cells/ml)"),
                average_total_cell_diameter=data.get(
                    float, "Estimated cell diameter (um)"
                ),
            )
        ],
    )



# THIS IS AN AUTOGENERATED FILE. DO NOT EDIT THIS FILE DIRECTLY.
from dataclasses import dataclass

UNITLESS = "(unitless)"


@dataclass(frozen=True, kw_only=True)
class HasUnit:
    unit: str


@dataclass(frozen=True, kw_only=True)
class Cell(HasUnit):
    unit: str = "cell"


@dataclass(frozen=True, kw_only=True)
class Centimeter(HasUnit):
    unit: str = "cm"


@dataclass(frozen=True, kw_only=True)
class Counts(HasUnit):
    unit: str = "Counts"


@dataclass(frozen=True, kw_only=True)
class CountsPerMicroliter(HasUnit):
    unit: str = "Counts/μL"


@dataclass(frozen=True, kw_only=True)
class CountsPerMilliliter(HasUnit):
    unit: str = "Counts/mL"


@dataclass(frozen=True, kw_only=True)
class CountsPerSecond(HasUnit):
    unit: str = "Counts/s"


@dataclass(frozen=True, kw_only=True)
class CountsTimesSecond(HasUnit):
    unit: str = "Counts.s"


@dataclass(frozen=True, kw_only=True)
class CubicMillimeter(HasUnit):
    unit: str = "mm^3"


@dataclass(frozen=True, kw_only=True)
class DegreeCelsius(HasUnit):
    unit: str = "degC"


@dataclass(frozen=True, kw_only=True)
class GramPerLiter(HasUnit):
    unit: str = "g/L"


@dataclass(frozen=True, kw_only=True)
class Hertz(HasUnit):
    unit: str = "Hz"


@dataclass(frozen=True, kw_only=True)
class KiloDalton(HasUnit):
    unit: str = "kDa"


@dataclass(frozen=True, kw_only=True)
class MicrogramPerMicroliter(HasUnit):
    unit: str = "ug/µL"


@dataclass(frozen=True, kw_only=True)
class MicrogramPerMilliliter(HasUnit):
    unit: str = "ug/mL"


@dataclass(frozen=True, kw_only=True)
class Microliter(HasUnit):
    unit: str = "μL"


@dataclass(frozen=True, kw_only=True)
class MicroliterPerMinute(HasUnit):
    unit: str = "µL/min"


@dataclass(frozen=True, kw_only=True)
class Micrometer(HasUnit):
    unit: str = "µm"


@dataclass(frozen=True, kw_only=True)
class MilliAbsorbanceUnit(HasUnit):
    unit: str = "mAU"


@dataclass(frozen=True, kw_only=True)
class MilliAbsorbanceUnitTimesMilliliter(HasUnit):
    unit: str = "mAU.mL"


@dataclass(frozen=True, kw_only=True)
class MilliAbsorbanceUnitTimesSecond(HasUnit):
    unit: str = "mAU.s"


@dataclass(frozen=True, kw_only=True)
class MilliOsmolesPerKilogram(HasUnit):
    unit: str = "mosm/kg"


@dataclass(frozen=True, kw_only=True)
class MilliSecond(HasUnit):
    unit: str = "ms"


@dataclass(frozen=True, kw_only=True)
class MilligramPerLiter(HasUnit):
    unit: str = "mg/L"


@dataclass(frozen=True, kw_only=True)
class MilligramPerMilliliter(HasUnit):
    unit: str = "mg/mL"


@dataclass(frozen=True, kw_only=True)
class Milliliter(HasUnit):
    unit: str = "mL"


@dataclass(frozen=True, kw_only=True)
class MilliliterPerLiter(HasUnit):
    unit: str = "mL/L"


@dataclass(frozen=True, kw_only=True)
class Millimeter(HasUnit):
    unit: str = "mm"


@dataclass(frozen=True, kw_only=True)
class MillimeterOfMercury(HasUnit):
    unit: str = "mmHg"


@dataclass(frozen=True, kw_only=True)
class MillimolePerLiter(HasUnit):
    unit: str = "mmol/L"


@dataclass(frozen=True, kw_only=True)
class MillionCellsPerMilliliter(HasUnit):
    unit: str = "10^6 cells/mL"


@dataclass(frozen=True, kw_only=True)
class Millivolt(HasUnit):
    unit: str = "mV"


@dataclass(frozen=True, kw_only=True)
class MillivoltTimesSecond(HasUnit):
    unit: str = "mV.s"


@dataclass(frozen=True, kw_only=True)
class Molar(HasUnit):
    unit: str = "M"


@dataclass(frozen=True, kw_only=True)
class NanoCoulomb(HasUnit):
    unit: str = "nC"


@dataclass(frozen=True, kw_only=True)
class NanoCoulombTimesSecond(HasUnit):
    unit: str = "nC.s"


@dataclass(frozen=True, kw_only=True)
class NanogramPerMicroliter(HasUnit):
    unit: str = "ng/µL"


@dataclass(frozen=True, kw_only=True)
class NanogramPerMilliliter(HasUnit):
    unit: str = "ng/mL"


@dataclass(frozen=True, kw_only=True)
class Nanometer(HasUnit):
    unit: str = "nm"


@dataclass(frozen=True, kw_only=True)
class Nanomolar(HasUnit):
    unit: str = "nM"


@dataclass(frozen=True, kw_only=True)
class Number(HasUnit):
    unit: str = "#"


@dataclass(frozen=True, kw_only=True)
class NumberPerMicroliter(HasUnit):
    unit: str = "#/μL"


@dataclass(frozen=True, kw_only=True)
class OpticalDensity(HasUnit):
    unit: str = "OD"


@dataclass(frozen=True, kw_only=True)
class PH(HasUnit):
    unit: str = "pH"


@dataclass(frozen=True, kw_only=True)
class PartsPerBillion(HasUnit):
    unit: str = "ppb"


@dataclass(frozen=True, kw_only=True)
class PartsPerMillion(HasUnit):
    unit: str = "ppm"


@dataclass(frozen=True, kw_only=True)
class PerSecond(HasUnit):
    unit: str = "s^-1"


@dataclass(frozen=True, kw_only=True)
class Percent(HasUnit):
    unit: str = "%"


@dataclass(frozen=True, kw_only=True)
class PicoAmpere(HasUnit):
    unit: str = "pA"


@dataclass(frozen=True, kw_only=True)
class PicoAmpereTimesSecond(HasUnit):
    unit: str = "pA.s"


@dataclass(frozen=True, kw_only=True)
class PicogramPerMilliliter(HasUnit):
    unit: str = "pg/mL"


@dataclass(frozen=True, kw_only=True)
class RelativeFluorescenceUnit(HasUnit):
    unit: str = "RFU"


@dataclass(frozen=True, kw_only=True)
class RelativeFluorescenceUnitTimesMilliliter(HasUnit):
    unit: str = "RFU.mL"


@dataclass(frozen=True, kw_only=True)
class RelativeFluorescenceUnitTimesSecond(HasUnit):
    unit: str = "RFU.s"


@dataclass(frozen=True, kw_only=True)
class RelativeLightUnit(HasUnit):
    unit: str = "RLU"


@dataclass(frozen=True, kw_only=True)
class RelativeLightUnitTimesMilliliter(HasUnit):
    unit: str = "RLU.mL"


@dataclass(frozen=True, kw_only=True)
class RelativeLightUnitTimesSecond(HasUnit):
    unit: str = "RLU.s"


@dataclass(frozen=True, kw_only=True)
class SecondTime(HasUnit):
    unit: str = "s"


@dataclass(frozen=True, kw_only=True)
class SquareCentimetersPerGram(HasUnit):
    unit: str = "cm^2/g"


@dataclass(frozen=True, kw_only=True)
class SquareCentimetersPerMole(HasUnit):
    unit: str = "cm^2/mol"


@dataclass(frozen=True, kw_only=True)
class TODO(HasUnit):
    unit: str = "TODO"


@dataclass(frozen=True, kw_only=True)
class UnitPerLiter(HasUnit):
    unit: str = "U/L"


@dataclass(frozen=True, kw_only=True)
class Unitless(HasUnit):
    unit: str = UNITLESS



from allotropy.allotrope.models.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.benchling_empower.benchling_empower_reader import (
    BenchlingEmpowerReader,
)
from allotropy.parsers.benchling_empower.benchling_empower_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.benchling_empower.constants import DISPLAY_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class BenchlingEmpowerParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = BenchlingEmpowerReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = BenchlingEmpowerReader(named_file_contents)

        return Data(
            create_metadata(
                reader.metadata_fields,
                reader.injections[0],
                named_file_contents.original_file_path,
            ),
            create_measurement_groups(reader.injections),
        )



from dataclasses import dataclass
from typing import Any, TypeVar

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import (
    CalculatedDataDocumentItem,
    ContainerType,
    DataProcessingDocument,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    DPCRAggregateDocument,
    DPCRDocumentItem,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    PassiveReferenceDyeDataCube,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    ReporterDyeDataCube,
    SampleDocument,
    TCalculatedDataAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueNumber,
    TQuantityValueNumberPerMicroliter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import TQuantityValue
from allotropy.allotrope.schema_mappers.data_cube import DataCube, get_data_cube
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class Error:
    error: str
    error_feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    sample_identifier: str
    location_identifier: str
    measurement_time: str
    target_identifier: str

    # Measurements
    concentration: float
    positive_partition_count: float
    total_partition_count: float
    negative_partition_count: float | None = None
    confidence_interval__95__: float | None = None
    reporter_dye_data_cube: DataCube | None = None
    passive_reference_dye_data_cube: DataCube | None = None

    # Optional metadata
    sample_role_type: str | None = None
    plate_identifier: str | None = None

    # Optional settings
    reporter_dye_setting: str | None = None
    passive_reference_dye_setting: str | None = None
    flourescence_intensity_threshold_setting: float | None = None

    # error documents
    errors: list[Error] | None = None

    # custom
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: float
    experimental_data_identifier: str | None = None
    # error documents
    errors: list[Error] | None = None


@dataclass(frozen=True)
class Metadata:
    device_identifier: str
    device_type: str
    software_name: str
    model_number: str | None = None
    detection_type: str | None = None
    unc_path: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None
    file_name: str | None = None
    data_system_instance_id: str | None = None
    analyst: str | None = None
    measurement_time: str | None = None
    brand_name: str | None = None
    container_type: ContainerType | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


CubeClass = TypeVar("CubeClass")


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/pcr/BENCHLING/2023/09/dpcr.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            dPCR_aggregate_document=DPCRAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.metadata.device_identifier,
                    brand_name=data.metadata.brand_name,
                    product_manufacturer=data.metadata.product_manufacturer,
                ),
                data_system_document=DataSystemDocument(
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                dPCR_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data.calculated_data
                ),
            )
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> DPCRDocumentItem:
        return DPCRDocumentItem(
            measurement_aggregate_document=MeasurementAggregateDocument(
                experimental_data_identifier=measurement_group.experimental_data_identifier,
                plate_well_count=TQuantityValueNumber(
                    value=measurement_group.plate_well_count
                ),
                error_aggregate_document=self._get_error_aggregate_document(
                    measurement_group.errors
                ),
                container_type=metadata.container_type,
                measurement_document=[
                    self._get_measurement_document(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            )
        )

    def _get_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocumentItem:
        measurement_doc = MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.measurement_time),
            target_DNA_description=measurement.target_identifier,
            total_partition_count=TQuantityValueNumber(
                value=measurement.total_partition_count
            ),
            reporter_dye_data_cube=get_data_cube(
                measurement.reporter_dye_data_cube, ReporterDyeDataCube
            ),
            passive_reference_dye_data_cube=get_data_cube(
                measurement.passive_reference_dye_data_cube, PassiveReferenceDyeDataCube
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.errors
            ),
            sample_document=SampleDocument(
                sample_identifier=measurement.sample_identifier,
                well_location_identifier=measurement.location_identifier,
                well_plate_identifier=measurement.plate_identifier,
                sample_role_type=measurement.sample_role_type,
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        # TODO(nstender): I think these should prob be removed for now, since every example
                        # we have so far is constant over the whole doc (or at least moved to a nullable
                        # per-measurement metadata for overrides). Going with being consistent/adding more
                        # info than needed for now.
                        device_type=metadata.device_type,
                        device_identifier=metadata.device_identifier,
                        reporter_dye_setting=measurement.reporter_dye_setting,
                        passive_reference_dye_setting=measurement.passive_reference_dye_setting,
                    )
                ]
            ),
            processed_data_aggregate_document=ProcessedDataAggregateDocument(
                processed_data_document=[
                    ProcessedDataDocumentItem(
                        number_concentration=TQuantityValueNumberPerMicroliter(
                            value=measurement.concentration
                        ),
                        positive_partition_count=TQuantityValueNumber(
                            value=measurement.positive_partition_count
                        ),
                        negative_partition_count=quantity_or_none(
                            TQuantityValueNumber, measurement.negative_partition_count
                        ),
                        confidence_interval__95__=quantity_or_none(
                            TQuantityValueNumber, measurement.confidence_interval__95__
                        ),
                        data_processing_document=DataProcessingDocument(
                            flourescence_intensity_threshold_setting=TQuantityValueUnitless(
                                value=measurement.flourescence_intensity_threshold_setting
                            )
                        )
                        if measurement.flourescence_intensity_threshold_setting
                        else None,
                    )
                ]
            ),
        )
        return add_custom_information_document(measurement_doc, measurement.custom_info)

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDataItem] | None
    ) -> TCalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return TCalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    calculated_datum=TQuantityValue(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None
        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.error_feature)
                for error in errors
            ]
        )



import numpy as np
import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import determine_encoding
from allotropy.parsers.utils.pandas import read_csv


class QiacuitydPCRReader:
    SUPPORTED_EXTENSIONS = "csv"
    well_data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents):
        # Read in the file, skip first row since it does not have data in it.
        contents = named_file_contents.contents.read()
        encoding = (
            determine_encoding(contents, named_file_contents.encoding)
            if isinstance(contents, bytes)
            else None
        )
        named_file_contents.contents.seek(0)
        qiacuity_dpcr_data = read_csv(
            filepath_or_buffer=named_file_contents.contents, header=1, encoding=encoding
        ).replace(np.nan, None)
        qiacuity_dpcr_data.columns = qiacuity_dpcr_data.columns.str.replace("�", "μ")
        qiacuity_dpcr_data.columns = qiacuity_dpcr_data.columns.str.replace("Âμ", "μ")
        column_names = qiacuity_dpcr_data.columns.tolist()
        # Rename the blank column to specify that it's the Well Name column
        column_names[0] = "Well Name"
        column_index = pd.Index(column_names)
        qiacuity_dpcr_data.columns = column_index
        self.well_data = qiacuity_dpcr_data



"""" Reader file for Thermo Fisher Qubit Flex Parser"""

import numpy as np
import pandas as pd

from allotropy.constants import DEFAULT_ENCODING
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_csv, read_excel


class ThermoFisherQubitFlexReader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"

    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        """
        Reads the content of the provided named file and returns it as a pandas DataFrame.

        Parameters:
        named_file_contents (NamedFileContents): The named file contents to read.
            It includes the original file name, file contents, and encoding.

        Returns:
        pd.DataFrame: The content of the file as a pandas DataFrame.

        Raises:
        AllotropeConversionError: If the file format is not supported.
        """
        if named_file_contents.extension == "csv":
            df = read_csv(
                named_file_contents.contents,
                index_col=False,
                encoding=DEFAULT_ENCODING,
            )
        else:
            df = read_excel(named_file_contents.contents)
        df = df.replace(np.nan, None)
        return df



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
import xml.etree.ElementTree as Et

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.allotrope.schema_mappers.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    Analyte,
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import get_key_or_error
from allotropy.parsers.biorad_bioplex_manager import constants
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    num_to_chars,
    try_float,
    try_float_or_none,
    try_int,
)
from allotropy.parsers.utils.xml import (
    get_attrib_from_xml,
    get_children_with_tag,
    get_element_from_xml,
    get_val_from_xml,
    get_val_from_xml_or_none,
)


@dataclass(frozen=True)
class AnalyteMetadata:
    name: str
    region: int
    error_msg: str | None

    @property
    def error(self) -> Error | None:
        if not self.error_msg:
            return None
        return Error(error=self.error_msg, feature=self.name)

    @staticmethod
    def create(analyte_xml: Et.Element) -> AnalyteMetadata:
        error_code = get_attrib_from_xml(analyte_xml, "Reading", "Code")
        return AnalyteMetadata(
            name=get_val_from_xml(analyte_xml, "AnalyteName"),
            region=try_int(analyte_xml.attrib["RegionNumber"], "analyte_region"),
            error_msg=get_key_or_error(
                "error code", error_code, constants.ERROR_MAPPING
            )
            if error_code != "0"
            else None,
        )


@dataclass
class SampleMetadata:
    sample_type: SampleRoleType
    sample_identifier: str
    description: str | None
    sample_dilution: float | None
    errors: list[Error]
    analyte_region_dict: dict[str, str]

    @staticmethod
    def create(
        sample_metadata: Et.Element,
        member_well: Et.Element,
    ) -> SampleMetadata:
        analyte_metadata = [
            AnalyteMetadata.create(analyte_xml)
            for analyte_xml in get_element_from_xml(member_well, "MWAnalytes")
        ]

        return SampleMetadata(
            sample_type=get_key_or_error(
                "sample role type",
                sample_metadata.tag,
                constants.SAMPLE_ROLE_TYPE_MAPPING,
            ),
            sample_identifier=get_val_from_xml(sample_metadata, "Label"),
            description=get_val_from_xml_or_none(sample_metadata, "Description"),
            errors=[analyte.error for analyte in analyte_metadata if analyte.error],
            sample_dilution=try_float_or_none(
                get_val_from_xml_or_none(sample_metadata, "Dilution")
            ),
            analyte_region_dict={
                str(analyte.region): analyte.name for analyte in analyte_metadata
            },
        )

    @staticmethod
    def create_samples(samples_xml: Et.Element) -> dict[str, SampleMetadata]:
        return {
            get_well_name(member_well.attrib): SampleMetadata.create(
                sample_metadata, member_well
            )
            for sample_types in samples_xml
            for sample_metadata in sample_types
            for member_well_group in get_children_with_tag(
                sample_metadata, "MemberWells"
            )
            for member_well in member_well_group
        }


@dataclass
class Well:
    name: str
    sample_volume_setting: float
    detector_gain_setting: str
    minimum_assay_bead_count_setting: int
    well_total_events: int
    acquisition_time: str
    analyst: str
    xml: Et.Element

    @staticmethod
    def create(well_xml: Et.Element) -> Well:
        return Well(
            name=get_well_name(well_xml.attrib),
            sample_volume_setting=try_float(
                get_val_from_xml(well_xml, "RunSettings", "SampleVolume"),
                "sample_volume",
            ),
            detector_gain_setting=get_val_from_xml(
                well_xml, "RunConditions", "RP1Gain"
            ),
            minimum_assay_bead_count_setting=try_int(
                get_attrib_from_xml(
                    well_xml, "RunSettings", "BeadCount", "StopReadingCriteria"
                ),
                "minimum_assay_bead_count_settings",
            ),
            acquisition_time=get_val_from_xml(well_xml, "AcquisitionTime"),
            well_total_events=try_int(
                get_val_from_xml(well_xml, "TotalEvents"), "well_total_events"
            ),
            analyst=get_val_from_xml(well_xml, "User"),
            xml=well_xml,
        )


def create_analyte(
    bead_region_xml: Et.Element,
    analyte_region_dict: dict[str, str],
) -> Analyte:
    # Look up analyte name from sample
    assay_bead_identifier = bead_region_xml.attrib["RegionNumber"]
    return Analyte(
        identifier=random_uuid_str(),
        name=analyte_region_dict[assay_bead_identifier],
        value=try_float(get_val_from_xml(bead_region_xml, "Median"), "fluorescence"),
        assay_bead_identifier=assay_bead_identifier,
        assay_bead_count=try_int(
            get_val_from_xml(bead_region_xml, "RegionCount"), "assay_bead_count"
        ),
        statistic_datum_role=TStatisticDatumRole.median_role,
    )


def create_analytes(
    well_xml: Et.Element,
    analyte_region_dict: dict[str, str],
    regions_of_interest: list[str],
) -> list[Analyte]:
    return [
        create_analyte(bead, analyte_region_dict)
        for child in get_children_with_tag(well_xml, "BeadRegions")
        for bead in child
        if bead.attrib["RegionNumber"] in regions_of_interest
    ]


@dataclass
class SystemMetadata:
    serial_number: str
    controller_version: str
    analytical_method: str
    plate_id: str
    regions_of_interest: list[str] = field(default_factory=list)

    @staticmethod
    def create(xml_well: Et.Element) -> SystemMetadata:
        return SystemMetadata(
            serial_number=get_val_from_xml(xml_well, "MachineInfo", "SerialNumber"),
            controller_version=get_val_from_xml(
                xml_well, "MachineInfo", "MicroControllerVersion"
            ),
            analytical_method=get_val_from_xml(xml_well, "RunProtocolDocumentLocation"),
            plate_id=get_val_from_xml(xml_well, "PlateID"),
            regions_of_interest=[
                str(region.attrib["RegionNumber"])
                for region in get_element_from_xml(
                    xml_well, "RunSettings", "RegionsOfInterest"
                )
            ],
        )


def get_well_name(well_attrib: dict[str, str]) -> str:
    row_name = num_to_chars(try_int(well_attrib["RowNo"], "row_number") - 1)
    col_name = str(well_attrib["ColNo"])
    return f"{row_name}{col_name}"


def create_metadata(
    root_xml: Et.Element, system_metadata: SystemMetadata, file_path: str
) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        software_name=constants.SOFTWARE_NAME,
        software_version=root_xml.attrib["BioPlexManagerVersion"],
        equipment_serial_number=system_metadata.serial_number,
        firmware_version=system_metadata.controller_version,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        device_type=constants.DEVICE_TYPE,
    )


def create_measurement_group(
    well: Well,
    sample: SampleMetadata,
    system_metadata: SystemMetadata,
    experimental_data_id: str | None,
    experiment_type: str | None,
    plate_well_count: int,
) -> MeasurementGroup:
    return MeasurementGroup(
        experiment_type=experiment_type,
        plate_well_count=plate_well_count,
        analytical_method_identifier=system_metadata.analytical_method,
        experimental_data_identifier=experimental_data_id,
        analyst=well.analyst,
        container_type=constants.CONTAINER_TYPE,
        measurements=[
            Measurement(
                identifier=random_uuid_str(),
                measurement_time=well.acquisition_time,
                assay_bead_count=well.well_total_events,
                description=sample.description,
                sample_identifier=sample.sample_identifier,
                location_identifier=well.name,
                sample_role_type=sample.sample_type,
                well_plate_identifier=system_metadata.plate_id,
                sample_volume_setting=well.sample_volume_setting,
                dilution_factor_setting=sample.sample_dilution,
                detector_gain_setting=well.detector_gain_setting,
                minimum_assay_bead_count_setting=well.minimum_assay_bead_count_setting,
                analytes=create_analytes(
                    well.xml,
                    sample.analyte_region_dict,
                    system_metadata.regions_of_interest,
                ),
                errors=sample.errors,
            )
        ],
    )



# generated by datamodel-codegen:
#   filename:  multi-analyte-profiling.schema.json
#   timestamp: 2024-06-04T03:30:34+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.components.plate_reader import (
    SampleRoleType,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDateTimeStampValue,
    TQuantityValue,
    TStringValue,
)


@dataclass(kw_only=True)
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue
    field_index: int | None = None


@dataclass(kw_only=True)
class DataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass(kw_only=True)
class Manifest:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: str | None = None
    field_type: str | None = None
    shapes: list[str] | None = None


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_identifier: TStringValue | None = None
    device_type: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class ReferenceMaterialDocument:
    reference_material_identifier: TStringValue | None = None
    batch_identifier: TStringValue | None = None
    expiry_time_prescription: TDateTimeStampValue | None = None


@dataclass(kw_only=True)
class CalibrationResultDocumentItem:
    calibration_result_name: TStringValue | None = None
    calibration_result: TQuantityValueUnitless | None = None


@dataclass(kw_only=True)
class CalibrationResultAggregateDocument:
    calibration_result_document: list[CalibrationResultDocumentItem] | None = None


@dataclass(kw_only=True)
class CalibrationDocumentItem:
    calibration_name: TStringValue | None = None
    calibration_description: TStringValue | None = None
    calibration_time: TDateTimeStampValue | None = None
    expiry_time_prescription: TDateTimeStampValue | None = None
    calibration_report: TStringValue | None = None
    reference_material_document: ReferenceMaterialDocument | None = None
    calibration_result_aggregate_document: CalibrationResultAggregateDocument | None = (
        None
    )


@dataclass(kw_only=True)
class CalibrationAggregateDocument:
    calibration_document: list[CalibrationDocumentItem] | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue | None = None
    description: Any | None = None
    brand_name: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    device_document: list[DeviceDocumentItem] | None = None
    calibration_aggregate_document: CalibrationAggregateDocument | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    data_system_instance_identifier: TStringValue | None = None
    file_name: TStringValue | None = None
    UNC_path: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue
    description: Any | None = None
    batch_identifier: TStringValue | None = None
    group_identifier: TStringValue | None = None
    sample_role_type: SampleRoleType | None = None
    written_name: TStringValue | None = None
    location_identifier: TStringValue | None = None
    well_plate_identifier: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    sample_volume_setting: TQuantityValueMicroliter | None = None
    dilution_factor_setting: TQuantityValueUnitless | None = None
    detector_gain_setting: TStringValue | None = None
    minimum_assay_bead_count_setting: TQuantityValueNumber | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass(kw_only=True)
class AnalyteDocumentItem:
    analyte_identifier: TStringValue
    analyte_name: TStringValue
    assay_bead_identifier: TStringValue
    assay_bead_count: TQuantityValueNumber
    fluorescence: TQuantityValueRelativeFluorescenceUnit


@dataclass(kw_only=True)
class AnalyteAggregateDocument:
    analyte_document: list[AnalyteDocumentItem]


@dataclass(kw_only=True)
class ErrorDocumentItem:
    error: TStringValue
    error_feature: TStringValue | None = None


@dataclass(kw_only=True)
class ErrorAggregateDocument:
    error_document: list[ErrorDocumentItem] | None = None


@dataclass(kw_only=True)
class MeasurementDocumentItem:
    measurement_identifier: TStringValue
    measurement_time: TDateTimeStampValue
    sample_document: SampleDocument
    device_control_aggregate_document: DeviceControlAggregateDocument
    assay_bead_count: TQuantityValueNumber
    analyte_aggregate_document: AnalyteAggregateDocument
    error_aggregate_document: ErrorAggregateDocument | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocumentItem]
    analytical_method_identifier: TStringValue | None = None
    method_version: TStringValue | None = None
    experimental_data_identifier: TStringValue | None = None
    experiment_type: TStringValue | None = None
    container_type: TStringValue | None = None
    plate_well_count: TQuantityValueNumber | None = None


@dataclass(kw_only=True)
class MultiAnalyteProfilingDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument
    analyst: TStringValue | None = None
    submitter: TStringValue | None = None


@dataclass(kw_only=True)
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValue
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    calculated_data_identifier: TStringValue | None = None
    calculation_description: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class CalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass(kw_only=True)
class MultiAnalyteProfilingAggregateDocument:
    device_system_document: DeviceSystemDocument
    multi_analyte_profiling_document: list[MultiAnalyteProfilingDocumentItem]
    data_system_document: DataSystemDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None


@dataclass(kw_only=True)
class Model:
    field_asm_manifest: Manifest | str
    multi_analyte_profiling_aggregate_document: MultiAnalyteProfilingAggregateDocument | None = (
        None
    )



import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_absolute_q.constants import get_dye_settings
from allotropy.parsers.lines_reader import (
    CsvReader,
    EMPTY_STR_OR_CSV_LINE,
)
from allotropy.parsers.utils.values import assert_not_none

# we don't do chained assignment, disable to prevent spurious warning.
pd.options.mode.chained_assignment = None


class AppbioAbsoluteQReader:
    SUPPORTED_EXTENSIONS = "csv"
    data: pd.DataFrame
    common_columns: list[str]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        csv_reader = CsvReader.create(named_file_contents)
        df, self.common_columns = AppbioAbsoluteQReader.parse_dataframe(csv_reader)

        columns_to_rename = {}
        if "Name" in df and "Sample" not in df:
            columns_to_rename["Name"] = "Sample"
        if "Well ID" in df and "Well" not in df:
            columns_to_rename["Well ID"] = "Well"
        if "Run name" in df and "Run" not in df:
            columns_to_rename["Run name"] = "Run"

        if columns_to_rename:
            df = df.rename(columns=columns_to_rename)

        required_keys = {"Sample"}
        for key in required_keys:
            if key not in df:
                possible_keys = key
                if key in columns_to_rename:
                    possible_keys += f" or {columns_to_rename[key]}"
                msg = f"Input is missing required column '{possible_keys}'"
                raise AllotropeConversionError(msg)

        self.data = df.replace(np.nan, None)

    @staticmethod
    def parse_dataframe(csv_reader: CsvReader) -> tuple[pd.DataFrame, list[str]]:
        first_line = csv_reader.get()
        if not first_line:
            msg = "Can not parse empty file"
            raise AllotropeConversionError(msg)
        # If the first row only contains dye settings indicating where each dye setting section is,
        # this is a summary file.
        columns = first_line.split(",")
        dye_setting_columns = get_dye_settings(columns)
        empty_columns = [col for col in columns if not col]

        # If not a summary file, parse the csv as is
        if len(dye_setting_columns) + len(empty_columns) != len(columns):
            return (
                assert_not_none(
                    csv_reader.pop_csv_block_as_df(
                        EMPTY_STR_OR_CSV_LINE, header="infer"
                    ),
                    "Failed to parser dataframe from file.",
                ),
                [],
            )

        # Get rid of the dye section setting line and then read the rest of the csv
        # Pandas tries to rename columns with the same name in the header, so read without header
        # and then set the first row to columns and drop it.
        csv_reader.pop()
        df = assert_not_none(
            csv_reader.pop_csv_block_as_df(EMPTY_STR_OR_CSV_LINE, header=None),
            "Failed to parser dataframe from file.",
        )
        df.columns = pd.Index(df.iloc[0].astype(str).tolist())
        df = df[1:]

        # Get the start indices of the dye setting sections.
        dye_column_indices = [i for i, col in enumerate(columns) if col]
        dye_section_sizes = [
            dye_column_indices[i + 1] - dye_column_indices[i]
            for i in range(len(dye_column_indices) - 1)
        ]
        if not all(size == dye_section_sizes[0] for size in dye_section_sizes):
            msg = f"Expected length of dye sections to all be the same size, got: {dye_section_sizes}"
            raise AllotropeConversionError(msg)

        # The columns before the first and after the last dye setting section are common columns
        before_df = df.iloc[:, : dye_column_indices[0]]
        after_df = df.iloc[
            :,
            (dye_column_indices[0] + dye_section_sizes[0] * len(dye_setting_columns)) :,
        ]
        base_df = pd.concat([before_df, after_df], axis="columns")
        # The group column is left blank on a ffill basis.
        base_df.loc[:, "Group"] = base_df["Group"].ffill()

        # For each dye setting, pull the columns for the section out and combine with the common columns
        # to create a dataset that looks like non-summary files, with one dye setting per row.
        dye_dfs: list[pd.DataFrame] = []
        for dye_setting, start_index in zip(
            dye_setting_columns, dye_column_indices, strict=True
        ):
            dye_df = df.iloc[:, start_index : (start_index + dye_section_sizes[0])]
            dye_df.loc[:, "Dye"] = dye_setting
            dye_dfs.append(pd.concat([base_df, dye_df], axis="columns"))

        return pd.concat(dye_dfs, axis="index"), list(base_df.columns)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
import math
from pathlib import Path
import re

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    CalculatedDataItem,
    DataProcessing,
    DataSource,
    DistributionDocument,
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.beckman_pharmspec.constants import (
    DEVICE_TYPE,
    PHARMSPEC_SOFTWARE_NAME,
    REQUIRED_DISTRIBUTION_DOCUMENT_KEYS,
    UNIT_LOOKUP,
    VALID_CALCS,
)
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.utils.pandas import (
    map_rows,
    SeriesData,
)
from allotropy.parsers.utils.uuids import random_uuid_str

ZERO_FLOAT = 0.0


def _create_processed_data(data: SeriesData) -> DistributionDocument:

    return DistributionDocument(
        distribution_identifier=random_uuid_str(),
        particle_size=data[float, "Particle Size(µm)"],
        cumulative_count=data[float, "Cumulative Count"],
        cumulative_particle_density=data[float, "Cumulative Counts/mL"],
        differential_particle_density=data.get(
            float, "Differential Counts/mL", NEGATIVE_ZERO
        ),
        differential_count=data.get(float, "Differential Count", NEGATIVE_ZERO),
    )


@dataclass(frozen=True, kw_only=True)
class Distribution:
    name: str
    features: list[DistributionDocument]
    is_calculated: bool

    @staticmethod
    def create(df: pd.DataFrame, name: str) -> Distribution:
        return Distribution(
            name=name,
            features=map_rows(df, _create_processed_data),
            is_calculated=name in VALID_CALCS,
        )

    @staticmethod
    def create_distributions(df: pd.DataFrame) -> list[Distribution]:
        distributions = []
        for g, gdf in df.groupby("Run No."):
            distributions.append(Distribution.create(gdf, str(g)))
        return distributions


@dataclass(frozen=True, kw_only=True)
class Header:
    measurement_time: str
    flush_volume_setting: float
    detector_view_volume: float
    repetition_setting: int
    sample_volume_setting: float
    sample_identifier: str
    dilution_factor_setting: float
    analyst: str
    equipment_serial_number: str
    software_version: str

    @staticmethod
    def _get_software_version_report_string(report_string: str) -> str:
        match = re.search(r"v(\d+(?:\.\d+)?(?:\.\d+)?)", report_string)
        if match:
            return match.group(1)
        return "Unknown"

    @staticmethod
    def create(data: SeriesData) -> Header:
        return Header(
            measurement_time=data[str, "Sample Date"].replace(".", "-"),
            flush_volume_setting=0,
            detector_view_volume=data[float, "View Volume"],
            repetition_setting=data[int, "No Of Runs"],
            sample_volume_setting=data[float, "Sample Volume (mL)"],
            sample_identifier=data[str, "Probe"],
            dilution_factor_setting=data[float, "Dilution Factor"],
            analyst=data[str, "Operator Name"],
            software_version=Header._get_software_version_report_string(
                data.series.iloc[0]
            ),
            equipment_serial_number=data[str, "Sensor Serial Number"],
        )


def create_metadata(header: Header, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        unc_path=file_path,
        software_name=PHARMSPEC_SOFTWARE_NAME,
        software_version=header.software_version,
        equipment_serial_number=header.equipment_serial_number,
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_identifier=NOT_APPLICABLE,
        flush_volume_setting=header.flush_volume_setting,
        detector_view_volume=header.detector_view_volume,
        repetition_setting=header.repetition_setting,
        sample_volume_setting=header.sample_volume_setting,
        device_type=DEVICE_TYPE,
    )


def create_measurement_groups(
    header: Header, distributions: list[Distribution]
) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(
            analyst=header.analyst,
            measurements=[
                Measurement(
                    identifier=distribution.name,
                    measurement_time=header.measurement_time,
                    sample_identifier=header.sample_identifier,
                    data_processing=DataProcessing(
                        dilution_factor_setting=header.dilution_factor_setting,
                    ),
                    distribution_documents=distribution.features,
                    errors=[
                        Error(
                            error=NOT_APPLICABLE,
                            feature=f"{key.replace('_', ' ')} - {feature.distribution_identifier}",
                        )
                        for feature in distribution.features
                        for key in feature.__dict__.keys()
                        if key in REQUIRED_DISTRIBUTION_DOCUMENT_KEYS
                        and is_negative_zero(feature.__dict__[key])
                    ],
                )
                for distribution in [x for x in distributions if not x.is_calculated]
            ],
            data_processing_time=header.measurement_time,
        )
    ]


def create_calculated_data(
    distributions: list[Distribution],
) -> list[CalculatedDataItem]:
    particle_size_sources = defaultdict(list)
    for source in [
        feature
        for distribution in distributions
        if not distribution.is_calculated
        for feature in distribution.features
    ]:
        particle_size_sources[source.particle_size].append(source)

    return [
        CalculatedDataItem(
            identifier=random_uuid_str(),
            name=f"{distribution.name}_{name}".lower(),
            value=value,
            unit=UNIT_LOOKUP[name],
            data_sources=[
                DataSource(
                    identifier=x.distribution_identifier,
                    feature=name.replace("_", " "),
                )
                for x in particle_size_sources[feature.particle_size]
            ],
        )
        for distribution in distributions
        for feature in distribution.features
        if distribution.is_calculated
        # Ignore "distribution_identifier" attribute, and skip empty or -0.0 values
        for name, value in feature.__dict__.items()
        if name != "distribution_identifier"
        and value is not None
        and not is_negative_zero(value)
    ]


def is_negative_zero(x: float) -> bool:
    return x == ZERO_FLOAT and math.copysign(1, x) == -1



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    CalculatedDataItem,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
    ProcessedDataFeature,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.thermo_fisher_nanodrop_8000 import constants
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def read_absorbances(data: SeriesData) -> dict[float, float]:
    # Read absorbances from abs <idx>, nm <idx> pair columns.
    absorbances: dict[float, float] = {}
    # NOTE: this range is just a reasonable sanity check so we don't have to use a "while True"
    for i in range(1, len(data.series.index) + 1):
        if not data.has_key(f"nm {i}"):
            break
        wavelength = data.get(float, f"nm {i}", validate=SeriesData.NOT_NAN)
        absorbance = data.get(float, f"abs {i}", validate=SeriesData.NOT_NAN)
        if wavelength is None or absorbance is None:
            continue
        absorbances[wavelength] = absorbance

    # Read absorbance columns with format "a<wavelength> <suffix>?", handle these as they come up.
    if (a230_absorbance := data.get(float, "a230")) is not None:
        absorbances[230] = a230_absorbance
    if (a260_absorbance := data.get(float, "a260")) is not None:
        absorbances[260] = a260_absorbance
    if (a280_absorbance := data.get(float, ["a280", "a280 10mm"])) is not None:
        absorbances[280] = a280_absorbance

    return absorbances


def read_mass_concentration_capture_wavelength(
    data: SeriesData, experiment_type: str | None, absorbances: dict[float, float]
) -> float:
    # NOTE: mass concentration is captured by a different wavelength depending on the experiment type.
    # DNA and RNA are captured as 260nm, while other experiment types are typically 280nm.
    # Given this, we apply the following logic:
    #
    # capture concentration on the 260 measurement document if:
    #   - there is no experiment type and no 280 column
    # capture concentration on the 280 measurement document if:
    #   - the experiment type is something other than DNA or RNA
    #   - the experiment type is not specified
    mass_concentration_capture_wavelength = data.get(float, "cursor nm")
    if mass_concentration_capture_wavelength is None:
        is_na_experiment = experiment_type and "NA" in experiment_type
        if not is_na_experiment and absorbances.get(280) is not None:
            mass_concentration_capture_wavelength = 280
        else:
            mass_concentration_capture_wavelength = 260

    return mass_concentration_capture_wavelength


def create_calculated_data(
    data: SeriesData, measurements: list[Measurement]
) -> list[CalculatedDataItem]:
    # Read absorbance ratios from pre-calculated columns.
    absorbance_ratios = {}
    for numerator, denominator in constants.ABSORBANCE_RATIOS:
        ratio = data.get(
            float, [f"{numerator}/{denominator}", f"a{numerator}/a{denominator}"]
        )
        if ratio:
            absorbance_ratios[(numerator, denominator)] = ratio

    calculated_data = [
        CalculatedDataItem(
            identifier=random_uuid_str(),
            name=f"A{numerator}/{denominator}",
            value=ratio,
            unit=UNITLESS,
            data_sources=[
                DataSource(identifier=measurement.identifier, feature="absorbance")
                for measurement in measurements
                if measurement.detector_wavelength_setting in (numerator, denominator)
            ],
        )
        for (numerator, denominator), ratio in absorbance_ratios.items()
    ]

    # Read all formulas <idx> columns into calculated data.
    # NOTE: this range is just a reasonable sanity check so we don't have to use a "while True"
    for i in range(1, len(data.series.index) + 1):
        if not data.has_key(f"formula {i}") and data.has_key(f"formula name {i}"):
            break
        if (
            value := data.get(float, f"formula value {i}", validate=SeriesData.NOT_NAN)
        ) is None:
            continue

        calculated_data.append(
            CalculatedDataItem(
                identifier=random_uuid_str(),
                name=data[str, f"formula name {i}"],
                value=value,
                unit=UNITLESS,
                data_sources=[
                    DataSource(identifier=measurement.identifier, feature="absorbance")
                    for measurement in measurements
                    # Check if the measurement wavelengh is in the forumula, remove ".0" from int floats.
                    if str(measurement.detector_wavelength_setting).replace(".0", "")
                    in data[str, f"formula {i}"]
                ],
            )
        )

    return calculated_data


@dataclass
class SpectroscopyRow:
    analyst: str | None
    timestamp: str
    experiment_type: str | None
    measurements: list[Measurement]
    calculated_data: list[CalculatedDataItem]

    @staticmethod
    def create(data: SeriesData) -> SpectroscopyRow:
        absorbances = read_absorbances(data)
        experiment_type = data.get(str, "na type")
        mass_concentration_capture_wavelength = (
            read_mass_concentration_capture_wavelength(
                data, experiment_type, absorbances
            )
        )
        mass_concentration = get_first_not_none(
            lambda key: data.get(float, key),
            ["conc.", "conc", "concentration"],
        )
        unit = data.get(str, "units")

        measurements: list[Measurement] = []
        for wavelength, absorbance in absorbances.items():
            if absorbance is None:
                continue
            measurements.append(
                Measurement(
                    type_=MeasurementType.ULTRAVIOLET_ABSORBANCE,
                    identifier=random_uuid_str(),
                    absorbance=absorbance,
                    detector_wavelength_setting=wavelength,
                    sample_identifier=data.get(
                        str, "sample id", NOT_APPLICABLE, SeriesData.NOT_NAN
                    ),
                    well_plate_identifier=data.get(
                        str, "plate id", validate=SeriesData.NOT_NAN
                    ),
                    location_identifier=data[str, "well"],
                    processed_data=ProcessedData(
                        features=[
                            ProcessedDataFeature(
                                result=mass_concentration,
                                unit=unit,
                            )
                        ],
                    )
                    if mass_concentration_capture_wavelength == wavelength
                    and mass_concentration
                    and unit
                    else None,
                )
            )

        calculated_data = create_calculated_data(data, measurements)

        return SpectroscopyRow(
            data.get(str, "user id"),
            f'{data[str, "date"]} {data.get(str, "time")}',
            experiment_type,
            measurements,
            calculated_data,
        )


def create_metadata(file_path: str) -> Metadata:
    return Metadata(
        device_identifier=constants.DEVICE_IDENTIFIER,
        device_type=constants.DEVICE_TYPE,
        model_number=constants.MODEL_NUBMER,
        file_name=Path(file_path).name,
        unc_path=file_path,
    )


def create_measurement_group(row: SpectroscopyRow) -> MeasurementGroup:
    return MeasurementGroup(
        measurement_time=row.timestamp,
        analyst=row.analyst,
        experiment_type=row.experiment_type,
        measurements=row.measurements,
    )



from __future__ import annotations

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import (
    Data,
    Mapper,
    MeasurementGroup,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_reader import QiacuitydPCRReader
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_structure import (
    create_measurements,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.pandas import map_rows
from allotropy.parsers.vendor_parser import VendorParser


class QiacuitydPCRParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Qiacuity dPCR"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = QiacuitydPCRReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = QiacuitydPCRReader(named_file_contents)
        return Data(
            create_metadata(named_file_contents.original_file_path),
            measurement_groups=[
                MeasurementGroup(
                    measurements=map_rows(reader.well_data, create_measurements),
                    # TODO: Hardcoded plate well count to 0 since it's a required field
                    #  ASM will be modified to optional in future version
                    plate_well_count=0,
                )
            ],
        )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
import re

import numpy as np
import pandas as pd

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.thermo_skanit.constants import DEVICE_TYPE, SAMPLE_ROLE_MAPPINGS
from allotropy.parsers.utils.pandas import df_to_series_data, parse_header_row
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass(frozen=True)
class ThermoSkanItMetadata:
    @staticmethod
    def create_metadata(
        instrument_info_df: pd.DataFrame, general_info_df: pd.DataFrame, file_path: str
    ) -> Metadata:

        # Replace empty with "" so we can add label columns together.
        instrument_info_df = instrument_info_df.fillna("")
        # The labels for data is spread across the first two columns for some reason, combine them as index.
        # NOTE: This is an assumption that may not be true for future files
        # Combine the first two columns into a Series
        combined_series = instrument_info_df.iloc[:, 0] + instrument_info_df.iloc[:, 1]
        # Convert the Series to an Index
        instrument_info_df.index = pd.Index(combined_series)
        # Read data from the last column, this is where values are found
        instrument_info_data = df_to_series_data(instrument_info_df.T, index=-1)
        general_info_data = df_to_series_data(parse_header_row(general_info_df.T))
        software_info = general_info_data.get(str, "Report generated with SW version")
        # Regular expression to extract software and version
        pattern = r"(SkanIt Software.*?)(?=,)|(\b\d+\.\d+\.\d+\.\d+\b)"
        if software_info is not None:
            matches = re.findall(pattern, software_info)
            # Process matches to get a cleaner output
            software_name = matches[0][0].strip() if matches[0][0] else None
            version_number = (
                matches[1][1] if len(matches) > 1 and matches[1][1] else None
            )
        else:
            msg = "Unable to identify Software Name or Version from General information tab"
            raise AllotropyParserError(msg)

        return Metadata(
            device_identifier=instrument_info_data[str, "Name"],
            model_number=instrument_info_data[str, "Name"],
            software_name=software_name,
            software_version=version_number,
            file_name=Path(file_path).name,
            unc_path=file_path,
            equipment_serial_number=instrument_info_data.get(str, "Serial number"),
        )


@dataclass(frozen=True)
class AbsorbanceDataWell(Measurement):
    @staticmethod
    def create(
        well_location: str,
        abs_value: float,
        sample_name: str,
        detector_wavelength: float,
    ) -> Measurement:
        return Measurement(
            type_=MeasurementType.ULTRAVIOLET_ABSORBANCE,
            identifier=random_uuid_str(),
            sample_identifier=sample_name,
            location_identifier=well_location,
            sample_role_type=AbsorbanceDataWell.get_sample_role(sample_name),
            absorbance=abs_value,
            detector_wavelength_setting=detector_wavelength,
            device_type=DEVICE_TYPE,
        )

    @staticmethod
    def get_sample_role(sample_name: str) -> SampleRoleType:
        stripped_text = re.sub(r"\d+", "", sample_name)
        try:
            return SAMPLE_ROLE_MAPPINGS[stripped_text]
        except KeyError as err:
            msg = f"Unable to identify sample role from {sample_name}"
            raise AllotropyParserError(msg) from err


@dataclass(frozen=True)
class ThermoSkanItMeasurementGroups:
    @staticmethod
    def create(
        absorbance_sheet_df: pd.DataFrame,
        layout_definitions_df: pd.DataFrame,
        session_info_df: pd.DataFrame,
    ) -> list[MeasurementGroup]:
        (
            abs_df,
            name_df,
            wavelength,
        ) = ThermoSkanItMeasurementGroups.identify_abs_and_sample_dfs(
            absorbance_sheet_df
        )
        plate_well_count = ThermoSkanItMeasurementGroups.get_plate_well_count(
            layout_definitions_df
        )

        session_info_data = df_to_series_data(parse_header_row(session_info_df.T))
        session_name = session_info_data.get(str, "Session notes")
        exec_time = session_info_data[str, "Execution time"]

        meas_groups = []
        # Stack the DataFrame, creating a MultiIndex
        stacked = abs_df.stack()

        # Iterate through the MultiIndex series and unpack it correctly
        for well_letter, well_column in stacked.index:
            abs_well = AbsorbanceDataWell.create(
                well_location=well_letter + str(well_column),
                abs_value=stacked.loc[well_letter, well_column],
                sample_name=str(name_df.loc[well_letter, well_column]),
                detector_wavelength=wavelength,
            )
            meas_groups.append(
                MeasurementGroup(
                    measurements=[abs_well],
                    plate_well_count=plate_well_count,
                    measurement_time=exec_time,
                    experimental_data_identifier=session_name,
                )
            )
        return meas_groups

    @staticmethod
    def get_plate_well_count(layout_definitions_df: pd.DataFrame) -> int:
        # Combine all non-empty values in the row where "Plate template" is found
        plate_template_row = layout_definitions_df[
            layout_definitions_df.iloc[:, 0].str.contains(
                "Plate template", case=False, na=False
            )
        ]
        plate_template_value = (
            plate_template_row.iloc[0].dropna().str.cat(sep=" ")
            if not plate_template_row.empty
            else None
        )
        msg = (
            "Unable to identify plate well count from Layout definitions tab. "
            "Expected to find in Plate template description."
        )
        if plate_template_value is None:
            raise AllotropyParserError(msg)
        else:
            # Extract the integer from the combined string
            search_string = re.search(r"\d+", plate_template_value)
            if search_string:
                plate_template_number = int(search_string.group())
                return plate_template_number
            else:
                raise AllotropyParserError(msg)

    @staticmethod
    def _set_headers(df: pd.DataFrame) -> pd.DataFrame:
        # Set the first column (well letters) as the index
        df.set_index(df.columns[0], inplace=True)
        # Set the first row (well numbers) as the columns
        df = parse_header_row(df)
        # Cast row numbers to int (float first to handle decimals, e.g. 1.0)
        df.columns = df.columns.astype(float).astype(int)
        return df

    @staticmethod
    def identify_abs_and_sample_dfs(
        absorbance_sheet_df: pd.DataFrame,
    ) -> tuple[pd.DataFrame, pd.DataFrame, float]:
        # Initialize variables
        start_reading_abs = False
        start_reading_sample = False
        data_between_abs_and_blank = []
        data_between_sample_and_blank = []

        # Iterate through each row
        for _, row in absorbance_sheet_df.iterrows():
            if pd.notna(row.iloc[0]) and "Wavelength" in row.iloc[0]:
                match = re.search(r"Wavelength:\s*(\d{3})\s*nm", row.iloc[0])
                if match:
                    wavelength = float(match.group(1))
                else:
                    msg = "Unable to identify Wavelength (nm) from Absorbance tab"
                    raise AllotropyParserError(msg)
            if "Abs" in row.values.astype(str):  # Check if 'abs' is in the row
                start_reading_abs = True
            if "Sample" in row.values.astype(str):  # Check if 'sample' is in the row
                start_reading_sample = True
            if pd.notna(row.iloc[0]) and "Autoloading" in row.iloc[0]:
                break

            if start_reading_sample:
                data_between_sample_and_blank.append(row)
            elif start_reading_abs:
                data_between_abs_and_blank.append(row)

        # Create DataFrames with the collected data
        df_between_abs_and_blank = ThermoSkanItMeasurementGroups._set_headers(
            pd.DataFrame(data_between_abs_and_blank)
        )
        df_between_sample_and_blank = ThermoSkanItMeasurementGroups._set_headers(
            pd.DataFrame(data_between_sample_and_blank)
        )

        return (df_between_abs_and_blank, df_between_sample_and_blank, wavelength)


@dataclass(frozen=True)
class DataThermoSkanIt(Data):
    @staticmethod
    def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
        # Prevent pandas from silently downcasting values, to prevent future incompatibility.
        pd.set_option("future.no_silent_downcasting", True)  # noqa: FBT003
        df = df.replace(r"^\s*$", np.nan, regex=True)
        df = df.dropna(axis="index", how="all")
        df = df.dropna(axis="columns", how="all")
        return df

    @staticmethod
    def create(sheet_data: dict[str, pd.DataFrame], file_path: str) -> Data:
        for sheet_name, df in sheet_data.items():
            clean_df = DataThermoSkanIt._clean_dataframe(df)
            # NOTE: This assumes a single absorbance plate on a single tab
            if "Absorbance" in sheet_name:
                abs_df = clean_df
            elif "Session information" in sheet_name:
                session_df = clean_df
            elif "Instrument information" in sheet_name:
                inst_info_df = clean_df
            elif "Layout" in sheet_name:
                layout_df = clean_df
            elif "General" in sheet_name:
                general_df = clean_df

        metadata = ThermoSkanItMetadata.create_metadata(
            instrument_info_df=inst_info_df,
            general_info_df=general_df,
            file_path=file_path,
        )
        measurement_groups = ThermoSkanItMeasurementGroups.create(
            absorbance_sheet_df=abs_df,
            layout_definitions_df=layout_df,
            session_info_df=session_df,
        )

        return Data(
            metadata=metadata,
            measurement_groups=measurement_groups,
        )



from re import search

from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    Measurement,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.creator import (
    create_data_cube,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.transformations import (
    Transformation,
)
from allotropy.parsers.cytiva_unicorn.structure.static_docs import (
    StaticDocs,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)
from allotropy.parsers.utils.uuids import random_uuid_str


class UnicornMeasurement(Measurement):
    @classmethod
    def filter_curve(
        cls, curve_elements: list[StrictXmlElement], pattern: str
    ) -> StrictXmlElement:
        for element in curve_elements:
            if search(pattern, element.find("Name").get_text()):
                return element
        msg = f"Unable to find curve element with pattern {pattern}"
        raise AllotropeConversionError(msg)

    @classmethod
    def get_data_cube_handler(
        cls, handler: UnicornZipHandler, curve_element: StrictXmlElement
    ) -> UnicornZipHandler:
        names = ["CurvePoints", "CurvePoint", "BinaryCurvePointsFileName"]
        data_name = curve_element.recursive_find(names).get_text()
        return handler.get_zip_from_pattern(data_name)

    @classmethod
    def get_data_cube(
        cls,
        handler: UnicornZipHandler,
        curve: StrictXmlElement,
        data_cube_component: DataCubeComponent,
        transformation: Transformation | None = None,
    ) -> DataCube:
        return create_data_cube(
            cls.get_data_cube_handler(handler, curve),
            curve.find("Name").get_text(),
            data_cube_component,
            transformation,
        )

    @classmethod
    def get_measurement(
        cls,
        static_docs: StaticDocs,
        device_control_docs: list[DeviceControlDoc],
        chromatogram_data_cube: DataCube | None = None,
        processed_data_chromatogram_data_cube: DataCube | None = None,
        derived_column_pressure_data_cube: DataCube | None = None,
    ) -> Measurement:
        return Measurement(
            measurement_identifier=random_uuid_str(),
            chromatography_serial_num=static_docs.chromatography_serial_num,
            column_inner_diameter=static_docs.column_inner_diameter,
            chromatography_chemistry_type=static_docs.chromatography_chemistry_type,
            chromatography_particle_size=static_docs.chromatography_particle_size,
            injection_identifier=static_docs.injection_identifier,
            injection_time=static_docs.injection_time,
            autosampler_injection_volume_setting=static_docs.autosampler_injection_volume_setting,
            sample_identifier=static_docs.sample_identifier,
            batch_identifier=static_docs.batch_identifier,
            chromatogram_data_cube=chromatogram_data_cube,
            device_control_docs=device_control_docs,
            processed_data_chromatogram_data_cube=processed_data_chromatogram_data_cube,
            derived_column_pressure_data_cube=derived_column_pressure_data_cube,
        )



# THIS IS AN AUTOGENERATED FILE. DO NOT EDIT THIS FILE DIRECTLY.
from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.definitions import (
    TNullableQuantityValue,
    TQuantityValue,
)
from allotropy.allotrope.models.shared.definitions.units import (
    Cell,
    Centimeter,
    Counts,
    CountsPerMicroliter,
    CountsPerMilliliter,
    CountsPerSecond,
    CountsTimesSecond,
    CubicMillimeter,
    DegreeCelsius,
    GramPerLiter,
    Hertz,
    KiloDalton,
    MicrogramPerMicroliter,
    MicrogramPerMilliliter,
    Microliter,
    MicroliterPerMinute,
    Micrometer,
    MilliAbsorbanceUnit,
    MilliAbsorbanceUnitTimesMilliliter,
    MilliAbsorbanceUnitTimesSecond,
    MilligramPerLiter,
    MilligramPerMilliliter,
    Milliliter,
    MilliliterPerLiter,
    Millimeter,
    MillimeterOfMercury,
    MillimolePerLiter,
    MillionCellsPerMilliliter,
    MilliOsmolesPerKilogram,
    MilliSecond,
    Millivolt,
    MillivoltTimesSecond,
    Molar,
    NanoCoulomb,
    NanoCoulombTimesSecond,
    NanogramPerMicroliter,
    NanogramPerMilliliter,
    Nanometer,
    Nanomolar,
    Number,
    NumberPerMicroliter,
    OpticalDensity,
    PartsPerBillion,
    PartsPerMillion,
    Percent,
    PerSecond,
    PH,
    PicoAmpere,
    PicoAmpereTimesSecond,
    PicogramPerMilliliter,
    RelativeFluorescenceUnit,
    RelativeFluorescenceUnitTimesMilliliter,
    RelativeFluorescenceUnitTimesSecond,
    RelativeLightUnit,
    RelativeLightUnitTimesMilliliter,
    RelativeLightUnitTimesSecond,
    SecondTime,
    SquareCentimetersPerGram,
    SquareCentimetersPerMole,
    TODO,
    Unitless,
    UnitPerLiter,
)


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCell(Cell, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCell(Cell, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCentimeter(Centimeter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCentimeter(Centimeter, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCounts(Counts, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCounts(Counts, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCountsPerMicroliter(CountsPerMicroliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCountsPerMicroliter(
    CountsPerMicroliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCountsPerMilliliter(CountsPerMilliliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCountsPerMilliliter(
    CountsPerMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCountsPerSecond(CountsPerSecond, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCountsPerSecond(CountsPerSecond, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCountsTimesSecond(CountsTimesSecond, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCountsTimesSecond(
    CountsTimesSecond, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueCubicMillimeter(CubicMillimeter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueCubicMillimeter(CubicMillimeter, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueDegreeCelsius(DegreeCelsius, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueDegreeCelsius(DegreeCelsius, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueGramPerLiter(GramPerLiter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueGramPerLiter(GramPerLiter, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueHertz(Hertz, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueHertz(Hertz, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueKiloDalton(KiloDalton, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueKiloDalton(KiloDalton, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMicrogramPerMicroliter(MicrogramPerMicroliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMicrogramPerMicroliter(
    MicrogramPerMicroliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMicrogramPerMilliliter(MicrogramPerMilliliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMicrogramPerMilliliter(
    MicrogramPerMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMicroliter(Microliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMicroliter(Microliter, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMicroliterPerMinute(MicroliterPerMinute, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMicroliterPerMinute(
    MicroliterPerMinute, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMicrometer(Micrometer, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMicrometer(Micrometer, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilliAbsorbanceUnit(MilliAbsorbanceUnit, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilliAbsorbanceUnit(
    MilliAbsorbanceUnit, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilliAbsorbanceUnitTimesMilliliter(
    MilliAbsorbanceUnitTimesMilliliter, TQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilliAbsorbanceUnitTimesMilliliter(
    MilliAbsorbanceUnitTimesMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilliAbsorbanceUnitTimesSecond(
    MilliAbsorbanceUnitTimesSecond, TQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilliAbsorbanceUnitTimesSecond(
    MilliAbsorbanceUnitTimesSecond, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilliOsmolesPerKilogram(MilliOsmolesPerKilogram, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilliOsmolesPerKilogram(
    MilliOsmolesPerKilogram, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilliSecond(MilliSecond, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilliSecond(MilliSecond, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilligramPerLiter(MilligramPerLiter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilligramPerLiter(
    MilligramPerLiter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilligramPerMilliliter(MilligramPerMilliliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilligramPerMilliliter(
    MilligramPerMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilliliter(Milliliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilliliter(Milliliter, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMilliliterPerLiter(MilliliterPerLiter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMilliliterPerLiter(
    MilliliterPerLiter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMillimeter(Millimeter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMillimeter(Millimeter, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMillimeterOfMercury(MillimeterOfMercury, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMillimeterOfMercury(
    MillimeterOfMercury, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMillimolePerLiter(MillimolePerLiter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMillimolePerLiter(
    MillimolePerLiter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMillionCellsPerMilliliter(
    MillionCellsPerMilliliter, TQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMillionCellsPerMilliliter(
    MillionCellsPerMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMillivolt(Millivolt, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMillivolt(Millivolt, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMillivoltTimesSecond(MillivoltTimesSecond, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMillivoltTimesSecond(
    MillivoltTimesSecond, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueMolar(Molar, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueMolar(Molar, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNanoCoulomb(NanoCoulomb, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNanoCoulomb(NanoCoulomb, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNanoCoulombTimesSecond(NanoCoulombTimesSecond, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNanoCoulombTimesSecond(
    NanoCoulombTimesSecond, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNanogramPerMicroliter(NanogramPerMicroliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNanogramPerMicroliter(
    NanogramPerMicroliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNanogramPerMilliliter(NanogramPerMilliliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNanogramPerMilliliter(
    NanogramPerMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNanometer(Nanometer, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNanometer(Nanometer, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNanomolar(Nanomolar, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNanomolar(Nanomolar, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNumber(Number, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNumber(Number, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueNumberPerMicroliter(NumberPerMicroliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueNumberPerMicroliter(
    NumberPerMicroliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueOpticalDensity(OpticalDensity, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueOpticalDensity(OpticalDensity, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePH(PH, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePH(PH, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePartsPerBillion(PartsPerBillion, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePartsPerBillion(PartsPerBillion, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePartsPerMillion(PartsPerMillion, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePartsPerMillion(PartsPerMillion, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePerSecond(PerSecond, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePerSecond(PerSecond, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePercent(Percent, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePercent(Percent, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePicoAmpere(PicoAmpere, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePicoAmpere(PicoAmpere, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePicoAmpereTimesSecond(PicoAmpereTimesSecond, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePicoAmpereTimesSecond(
    PicoAmpereTimesSecond, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValuePicogramPerMilliliter(PicogramPerMilliliter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValuePicogramPerMilliliter(
    PicogramPerMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueRelativeFluorescenceUnit(RelativeFluorescenceUnit, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueRelativeFluorescenceUnit(
    RelativeFluorescenceUnit, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueRelativeFluorescenceUnitTimesMilliliter(
    RelativeFluorescenceUnitTimesMilliliter, TQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueRelativeFluorescenceUnitTimesMilliliter(
    RelativeFluorescenceUnitTimesMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueRelativeFluorescenceUnitTimesSecond(
    RelativeFluorescenceUnitTimesSecond, TQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueRelativeFluorescenceUnitTimesSecond(
    RelativeFluorescenceUnitTimesSecond, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueRelativeLightUnit(RelativeLightUnit, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueRelativeLightUnit(
    RelativeLightUnit, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueRelativeLightUnitTimesMilliliter(
    RelativeLightUnitTimesMilliliter, TQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueRelativeLightUnitTimesMilliliter(
    RelativeLightUnitTimesMilliliter, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueRelativeLightUnitTimesSecond(
    RelativeLightUnitTimesSecond, TQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueRelativeLightUnitTimesSecond(
    RelativeLightUnitTimesSecond, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueSecondTime(SecondTime, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueSecondTime(SecondTime, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueSquareCentimetersPerGram(SquareCentimetersPerGram, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueSquareCentimetersPerGram(
    SquareCentimetersPerGram, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueSquareCentimetersPerMole(SquareCentimetersPerMole, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueSquareCentimetersPerMole(
    SquareCentimetersPerMole, TNullableQuantityValue
):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueTODO(TODO, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueTODO(TODO, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueUnitPerLiter(UnitPerLiter, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueUnitPerLiter(UnitPerLiter, TNullableQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TQuantityValueUnitless(Unitless, TQuantityValue):
    pass


@dataclass(frozen=True, kw_only=True)
class TNullableQuantityValueUnitless(Unitless, TNullableQuantityValue):
    pass



from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ContainerType

DISPLAY_NAME = "Bio-Rad CFX Maestro"
DEVICE_TYPE = "qPCR"

PRODUCT_MANUFACTURER = "Bio-Rad"
SOFTWARE_NAME = "CFX Maestro"
CONTAINER_TYPE = ContainerType.well_plate



import numpy as np
import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_excel, split_dataframe
from allotropy.parsers.utils.values import assert_not_none


class TecanMagellanReader:
    SUPPORTED_EXTENSIONS = "xlsx"
    metadata_lines: list[str]
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        data = read_excel(named_file_contents.contents).replace(np.nan, None)
        data, metadata = split_dataframe(
            data,
            lambda row: row.astype(str).iloc[0].startswith("Date of measurement"),
            include_split_row=True,
        )
        # Files might contain rows with metadata that we will not use, we identify
        # those by checking that the Well positions have the right format
        data = data.dropna(subset=["Well positions"])
        data = data[data["Well positions"].str.match(r"[A-Z]+\d+")]

        # Fill empty plate info if there's a unique value
        if "Plate" in data:
            if len(unique := data["Plate"].dropna().unique()) == 1:
                data["Plate"] = unique[0]

        self.data = data
        self.metadata_lines = assert_not_none(metadata).iloc[:, 0].to_list()



from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_qubit4.thermo_fisher_qubit4_reader import (
    ThermoFisherQubit4Reader,
)
from allotropy.parsers.thermo_fisher_qubit_flex import constants
from allotropy.parsers.thermo_fisher_qubit_flex.thermo_fisher_qubit_flex_reader import (
    ThermoFisherQubitFlexReader,
)
from allotropy.parsers.thermo_fisher_qubit_flex.thermo_fisher_qubit_flex_structure import (
    create_data,
)
from allotropy.parsers.vendor_parser import VendorParser


class ThermoFisherQubitFlexParser(VendorParser[Data, Model]):
    DISPLAY_NAME = constants.DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = ThermoFisherQubit4Reader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        df = ThermoFisherQubitFlexReader.read(named_file_contents)
        return create_data(df, named_file_contents.original_file_path)



from __future__ import annotations

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import read_csv


class CFXMaestroReader:
    SUPPORTED_EXTENSIONS = "csv"
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        self.data = read_csv(named_file_contents.contents)



from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.liquid_handler.benchling._2024._11.liquid_handler import (
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    LiquidHandlerAggregateDocument,
    LiquidHandlerDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class Error:
    error: str
    feature: str | None = None


@dataclass(frozen=True)
class Device:
    identifier: str
    serial_number: str
    product_manufacturer: str
    device_type: str


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    measurement_time: str
    sample_identifier: str

    # Optional metadata
    source_plate: str | None = None
    source_well: str | None = None
    source_location: str | None = None
    destination_plate: str | None = None
    destination_well: str | None = None
    destination_location: str | None = None

    # Measurements
    aspiration_volume: float | None = None
    transfer_volume: float | None = None

    # Optional settings
    injection_volume_setting: float | None = None

    # Errors
    errors: list[Error] | None = None

    device_control_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    analyst: str
    measurements: list[Measurement]
    analytical_method_identifier: str | None = None


@dataclass(frozen=True)
class Metadata:
    asm_file_identifier: str
    data_system_instance_identifier: str
    file_name: str
    unc_path: str
    device_type: str
    software_name: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None
    devices: list[Device] | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/liquid-handler/BENCHLING/2024/11/liquid-handler.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            liquid_handler_aggregate_document=LiquidHandlerAggregateDocument(
                liquid_handler_document=[
                    self._get_technique_document(group, data.metadata)
                    for group in data.measurement_groups
                ],
                device_system_document=DeviceSystemDocument(
                    equipment_serial_number=data.metadata.equipment_serial_number,
                    product_manufacturer=data.metadata.product_manufacturer,
                    device_document=[
                        DeviceDocumentItem(
                            device_type=device.device_type,
                            device_identifier=device.identifier,
                            equipment_serial_number=device.serial_number,
                            product_manufacturer=device.product_manufacturer,
                        )
                        for device in data.metadata.devices
                    ]
                    if data.metadata.devices
                    else None,
                ),
                data_system_document=DataSystemDocument(
                    ASM_file_identifier=data.metadata.asm_file_identifier,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                    software_name=data.metadata.software_name,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> LiquidHandlerDocumentItem:
        return LiquidHandlerDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                analytical_method_identifier=measurement_group.analytical_method_identifier,
                measurement_document=[
                    self._get_measurement_document_item(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.measurement_time),
            sample_document=self._get_sample_document(measurement),
            aspiration_volume=quantity_or_none(
                TQuantityValueMicroliter, measurement.aspiration_volume
            ),
            transfer_volume=quantity_or_none(
                TQuantityValueMicroliter, measurement.transfer_volume
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    add_custom_information_document(
                        DeviceControlDocumentItem(
                            device_type=metadata.device_type,
                            injection_volume_setting=quantity_or_none(
                                TQuantityValueMicroliter,
                                measurement.injection_volume_setting,
                            ),
                        ),
                        measurement.device_control_custom_info,
                    )
                ]
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.errors
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
            source_well_location_identifier=measurement.source_well,
            source_well_plate_identifier=measurement.source_plate,
            source_location_identifier=measurement.source_location,
            destination_well_location_identifier=measurement.destination_well,
            destination_well_plate_identifier=measurement.destination_plate,
            destination_location_identifier=measurement.destination_location,
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None

        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.feature)
                for error in errors
            ]
        )



from __future__ import annotations

from dataclasses import dataclass
from io import StringIO
from pathlib import PureWindowsPath
import re

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    ImageFeature,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.ctl_immunospot import constants
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_or_nan,
)


@dataclass(frozen=True)
class Plate:
    name: str
    wells: dict[str, JsonFloat]

    @staticmethod
    def create(reader: LinesReader) -> Plate:
        raw_name = assert_not_none(
            reader.pop(),
            msg="Unable to read name of assay data.",
        )

        name_match = assert_not_none(
            re.search("\t\t\t(.+)", raw_name),
            msg="Unable to parse name of assay data.",
        )
        name = re.sub(r"\s+", " ", name_match.group(1))

        if "ImmunoSpot Plate Code" in name:
            name = "Spot Count"

        raw_columns = assert_not_none(
            reader.pop(),
            msg=f"Unable to read data from {name}",
        )
        columns = re.sub(r"\s+", " ", raw_columns.strip())

        reader.drop_empty()

        wells = {}
        for raw_line in reader.pop_until_empty():
            line = re.sub(r"\s+", " ", raw_line)
            if match := re.match(r"^([A-Z])\s+(.+)", line):
                raw_values = match.group(2).strip()
                for column, value in zip(
                    columns.split(), raw_values.split(), strict=True
                ):
                    wells[match.group(1) + column] = try_float_or_nan(value)

        return Plate(name, wells)


@dataclass(frozen=True)
class AssayData:
    plates: list[Plate]
    identifier: str | None

    @staticmethod
    def create(reader: LinesReader) -> AssayData:
        reader.drop_until_inclusive("Unprocessed Data$")

        reader.drop_empty()
        plate_code_line = reader.get() or ""
        assert_not_none(
            re.search("Plate Code =", plate_code_line),
            msg="Unable to find ImmunoSpot Plate Code line",
        )

        identifier = (
            match.group(1)
            if (match := re.search(r"Plate Code = ([\w ]+)", plate_code_line))
            else None
        )

        plates = []
        while reader.current_line_exists():
            reader.drop_empty()
            plates.append(Plate.create(reader))
            reader.drop_empty()

        if not plates:
            msg = "Unable to find plate information."
            raise AllotropeConversionError(msg)

        return AssayData(plates, identifier)


def create_measurement_groups(
    assay_data: AssayData, header: Header
) -> list[MeasurementGroup]:
    well_plate_identifier = (
        assay_data.identifier or PureWindowsPath(header.file_path).stem
    )
    plate_well_count = len(assay_data.plates[0].wells)
    return [
        MeasurementGroup(
            plate_well_count=plate_well_count,
            measurement_time=header.counted_time,
            analyst=header.authenticated_user,
            measurements=[
                Measurement(
                    type_=MeasurementType.OPTICAL_IMAGING,
                    device_type=constants.DEVICE_TYPE,
                    identifier=random_uuid_str(),
                    well_plate_identifier=well_plate_identifier,
                    location_identifier=well_position,
                    sample_identifier=f"{well_plate_identifier}_{well_position}",
                    detection_type=constants.DETECTION_TYPE,
                    processed_data=ProcessedData(
                        identifier=random_uuid_str(),
                        features=[
                            ImageFeature(
                                identifier=random_uuid_str(),
                                feature=plate.name,
                                result=plate.wells[well_position],
                            )
                            for plate in assay_data.plates
                        ],
                    ),
                )
            ],
        )
        for well_position in assay_data.plates[0].wells
    ]


@dataclass
class Header:
    counted_time: str
    file_path: str
    analyzer_serial_number: str
    software_version: str
    computer_name: str
    authenticated_user: str

    @staticmethod
    def create(reader: LinesReader) -> Header:
        lines = [
            # Add missing key for file path line.
            f"File path: {line}" if line.endswith(".txt") else line
            for raw_line in reader.pop_until_empty()
            # Split rows over ';'
            for line in raw_line.split(";")
        ]
        reader.drop_empty()
        lines.extend(list(reader.pop_until_empty()))

        df = read_csv(
            StringIO("\n".join(lines)),
            sep=r"^([^:]+):\s+",
            header=None,
            engine="python",
            index_col=1,
        ).T
        data = df_to_series_data(df, index=1)

        return Header(
            counted_time=data[str, "Counted"],
            file_path=data[str, "File path"],
            analyzer_serial_number=data[str, "Analyzer Serial number"],
            software_version=data[str, "Software version"],
            computer_name=data[str, "Computer name"],
            authenticated_user=data[str, "Authenticated user"],
        )


def create_metadata(header: Header) -> Metadata:
    return Metadata(
        file_name=PureWindowsPath(header.file_path).name,
        unc_path=header.file_path,
        device_identifier=NOT_APPLICABLE,
        model_number=assert_not_none(
            re.match(r"^(\w+)-(\w+)", header.analyzer_serial_number),
            msg="Unable to parse analyzer serial number.",
        ).group(1),
        data_system_instance_id=header.computer_name,
        equipment_serial_number=header.analyzer_serial_number,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        software_name=constants.SOFTWARE_NAME,
        software_version=assert_not_none(
            re.match(r"^ImmunoSpot ([\d\.]+)$", header.software_version),
            msg="Unable to parse software version",
        ).group(1),
    )



from allotropy.allotrope.models.adm.cell_counting.rec._2024._09.cell_counting import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.chemometec_nucleoview.nucleoview_reader import NucleoviewReader
from allotropy.parsers.chemometec_nucleoview.nucleoview_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.pandas import df_to_series_data, map_rows
from allotropy.parsers.vendor_parser import VendorParser


class ChemometecNucleoviewParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "ChemoMetec Nucleoview"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = NucleoviewReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        df = NucleoviewReader.read(named_file_contents.contents)
        return Data(
            create_metadata(
                df_to_series_data(df.head(1)), named_file_contents.original_file_path
            ),
            map_rows(df, create_measurement_groups),
        )



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

from dateutil import parser
import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat, NaN
from allotropy.allotrope.schema_mappers.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Analyte,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.roche_cedex_bioht.constants import (
    BELOW_TEST_RANGE,
    MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE,
    OPTICAL_DENSITY,
    SOLUTION_ANALYZER,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass(frozen=True)
class Title:
    data_processing_time: str
    analyst: str
    device_serial_number: str
    model_number: str | None = None
    software_version: str | None = None

    @staticmethod
    def create(title_data: SeriesData) -> Title:
        return Title(
            title_data[str, "data processing time"],
            title_data[str, "analyst"],
            title_data[str, "device serial number"],
            title_data.get(str, "model number"),
            title_data.get(str, "software version"),
        )


@dataclass(frozen=True)
class RawMeasurement:
    name: str
    measurement_time: str
    concentration_value: JsonFloat
    unit: str
    error: str | None = None

    @staticmethod
    def create(data: SeriesData) -> RawMeasurement:
        error = data.get(str, "flag", "").strip()
        # TODO: handle other errors
        if BELOW_TEST_RANGE not in error:
            error = ""
        concentration_value = (
            NaN
            if BELOW_TEST_RANGE in error
            else data.get(float, "concentration value", NaN)
        )
        return RawMeasurement(
            data[str, "analyte name"],
            data[str, "measurement time"],
            concentration_value,
            data[str, "concentration unit"],
            error or None,
        )


def create_measurements(data: pd.DataFrame) -> dict[str, dict[str, RawMeasurement]]:
    measurements = sorted(
        map_rows(data, RawMeasurement.create), key=lambda a: a.measurement_time
    )

    # Dict from measurement time to data
    groups: defaultdict[str, dict[str, RawMeasurement]] = defaultdict(dict)

    current_measurement_time = measurements[0].measurement_time
    previous_measurement_time = current_measurement_time
    for analyte in measurements:
        time_diff = parser.parse(analyte.measurement_time) - parser.parse(
            previous_measurement_time
        )
        if time_diff > MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE:
            current_measurement_time = analyte.measurement_time
        if analyte.name in groups[current_measurement_time]:
            if analyte.concentration_value is NaN:
                continue
            # NOTE: if this fails, it's probably because MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE is too big
            # and we're erroneously grouping two groups of measurements into one.
            # We could potentially make this more robust by just splitting into a new group if a duplicate
            # measurement is found, but cross that bridge when we come to it.
            if (
                groups[current_measurement_time][analyte.name].concentration_value
                is not NaN
            ):
                msg = f"Duplicate measurement for {analyte.name} in the same measurement group: {analyte.concentration_value} vs {groups[current_measurement_time][analyte.name].concentration_value}"
                raise AllotropyParserError(msg)
        groups[current_measurement_time][analyte.name] = analyte
        previous_measurement_time = analyte.measurement_time

    return dict(groups)


@dataclass(frozen=True)
class Sample:
    name: str
    measurements: dict[str, dict[str, RawMeasurement]]
    batch: str | None = None

    @staticmethod
    def create(name: str, batch: str | None, sample_data: pd.DataFrame) -> Sample:
        return Sample(
            name,
            create_measurements(sample_data),
            batch=batch or None,
        )

    @staticmethod
    def create_samples(samples_data: pd.DataFrame) -> list[Sample]:
        return [
            Sample.create(name, batch, samples_data)
            for (name, batch), samples_data in samples_data.groupby(
                # A sample group is defined by both the sample and the batch identifier
                ["sample identifier", "batch identifier"]
            )
        ]


def _create_measurements(
    sample: Sample, measurement_time: str, raw_measurements: dict[str, RawMeasurement]
) -> list[Measurement]:
    measurements: list[Measurement] = []

    analytes: list[Analyte] = []
    for name in sorted(raw_measurements):
        measurement = raw_measurements[name]
        value = measurement.concentration_value
        # TODO: report value and add error
        if value is NaN:
            continue

        if name == OPTICAL_DENSITY:
            measurements.append(
                Measurement(
                    identifier=random_uuid_str(),
                    measurement_time=measurement_time,
                    sample_identifier=sample.name,
                    batch_identifier=sample.batch,
                    absorbance=value if isinstance(value, float) else -1,
                )
            )
        else:
            analytes.append(
                Analyte(
                    name=measurement.name,
                    value=value if isinstance(value, float) else -1,
                    unit=measurement.unit,
                )
            )

    if analytes:
        measurements.append(
            Measurement(
                identifier=random_uuid_str(),
                measurement_time=measurement_time,
                sample_identifier=sample.name,
                batch_identifier=sample.batch,
                analytes=analytes,
            )
        )

    return measurements


def create_measurement_groups(
    samples: list[Sample], title: Title
) -> list[MeasurementGroup]:
    groups: list[MeasurementGroup] = []
    for sample in samples:
        measurements = [
            measurement
            for measurement_time, sample_measurements in sample.measurements.items()
            for measurement in _create_measurements(
                sample, measurement_time, sample_measurements
            )
        ]
        if not measurements:
            continue
        groups.append(
            MeasurementGroup(
                analyst=title.analyst,
                data_processing_time=title.data_processing_time,
                measurements=measurements,
            )
        )
    return groups


def create_metadata(title: Title, file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        unc_path=file_path,
        device_type=SOLUTION_ANALYZER,
        model_number=title.model_number,
        equipment_serial_number=title.device_serial_number,
        device_identifier=NOT_APPLICABLE,
        software_name=title.model_number,
        software_version=title.software_version,
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_identifier=NOT_APPLICABLE,
    )



from decimal import Decimal
from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.constants import DEFAULT_EPOCH_TIMESTAMP, NOT_APPLICABLE
from allotropy.parsers.revvity_matrix import constants
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        file_name=path.name,
        unc_path=file_path,
        device_type=constants.DEVICE_TYPE,
    )


def create_measurement_group(data: SeriesData) -> MeasurementGroup:
    # This function will be called for every row in the dataset, use it to create
    # a corresponding measurement group.

    # Cell counts are measured in cells/mL, but reported in millions of cells/mL
    viable_cell_density = float(
        Decimal(data[float, "Live Cells/mL"]) / Decimal("1000000")
    )
    total_cell_density = data.get(float, "Total Cells/mL")
    if total_cell_density:
        total_cell_density = float(Decimal(total_cell_density) / Decimal("1000000"))
    dead_cell_density = data.get(float, "Dead Cells/mL")
    if dead_cell_density:
        dead_cell_density = float(Decimal(dead_cell_density) / Decimal("1000000"))

    errors = data.get(str, "Errors:", validate=SeriesData.NOT_NAN)
    return MeasurementGroup(
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                # NOTE: instrument file does not provide a timestamp, but it is required by ASM, so pass
                # EPOCH to signal no timestamp.
                timestamp=DEFAULT_EPOCH_TIMESTAMP,
                sample_identifier=data[str, "Well Name"],
                viability=data[float, "Viability"],
                total_cell_count=data.get(float, "Total Count"),
                total_cell_density=total_cell_density,
                average_total_cell_diameter=data.get(float, "Total Mean Size"),
                viable_cell_count=data.get(float, "Live Count"),
                viable_cell_density=viable_cell_density,
                average_live_cell_diameter=data.get(float, "Live Mean Size"),
                dead_cell_count=data.get(float, "Dead Count"),
                dead_cell_density=dead_cell_density,
                average_dead_cell_diameter=data.get(float, "Dead Mean Size"),
                errors=[
                    Error(error=error)
                    for error in (errors.split(",") if errors else [])
                ],
            )
        ]
    )



from collections import defaultdict
from pathlib import Path
from typing import Any

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.liquid_chromatography.benchling._2023._09.liquid_chromatography import (
    DeviceControlDoc,
    Measurement,
    MeasurementGroup,
    Metadata,
    Peak,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.benchling_empower import constants
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none, try_float, try_float_or_none


def create_metadata(
    metadata: dict[str, Any], first_injection: dict[str, Any], file_path: str
) -> Metadata:
    return Metadata(
        asset_management_identifier=metadata.get("SystemName", NOT_APPLICABLE),
        analyst=metadata.get("SampleSetAcquiredBy", NOT_APPLICABLE),
        software_name=constants.SOFTWARE_NAME,
        software_version=first_injection.get("AcqSWVersion"),
        file_name=Path(file_path).name,
        unc_path=file_path,
        description=metadata.get("SystemComments", metadata.get("SampleSetComments")),
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
    )


def _get_chromatogram(injection: dict[str, Any]) -> DataCube | None:
    chrom: list[list[float]] | None = injection.get("chrom")
    if not chrom:
        return None
    if len(chrom) != 2:
        msg = "Expected chrom to have two lists"
        raise AllotropeConversionError(msg)

    dimensions, measures = chrom

    # ASM expects chromatogram absorbance in mAU, convert units if different.
    detection_unit = injection["DetUnits"]
    if detection_unit == "AU":
        measures = [m * 1000 for m in measures]
    elif detection_unit != "mAU":
        msg = f"Unexpected Chromatogram detection unit: {detection_unit}"
        raise AllotropeConversionError(msg)

    # ASM expected chromatogram dimensions (x axis) to be in seconds, but Empower reports it in minutes,
    # so convert here.
    dimensions = [t * 60 for t in dimensions]

    return DataCube(
        label="absorbance",
        structure_dimensions=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="retention time",
                unit="s",
            )
        ],
        structure_measures=[
            DataCubeComponent(
                type_=FieldComponentDatatype.double,
                concept="absorbance",
                unit="mAU",
            )
        ],
        dimensions=[dimensions],
        measures=[measures],
    )


def _create_peak(peak: dict[str, Any]) -> Peak:
    # Area and height are reported in μV, but are reported in ASM as mAU
    # For Empower software, 1V == 1AU, so we just need to convert μ to m
    if (area := try_float_or_none(peak.get("Area"))) is not None:
        area /= 1000
    if (height := try_float_or_none(peak.get("Height"))) is not None:
        height /= 1000
    # Times are reported in minutes by Empower - convert to seconds
    if (retention_time := try_float_or_none(peak.get("RetentionTime"))) is not None:
        retention_time *= 60

    return Peak(
        identifier=random_uuid_str(),
        # Times are reported in minutes by Empower - convert to seconds
        start=try_float(peak.get("StartTime"), "StartTime") * 60,
        start_unit="s",
        end=try_float(peak.get("EndTime"), "EndTime") * 60,
        end_unit="s",
        retention_time=retention_time,
        area=area,
        area_unit="mAU.s",
        relative_area=try_float_or_none(peak.get("PctArea")),
        width=try_float_or_none(peak.get("Width")),
        height=height,
        relative_height=try_float_or_none(peak.get("PctHeight")),
        written_name=peak.get("Name"),
    )


def _create_measurements(injection: dict[str, Any]) -> list[Measurement]:
    peaks: list[dict[str, Any]] = injection.get("peaks", [])
    sample_type = injection.get("SampleType")
    sample_role_type = None
    if sample_type:
        sample_role_types = [
            srt.value for srt in SampleRoleType if str(sample_type).lower() in srt.value
        ]
        sample_role_type = (
            sample_role_types[0]
            if len(sample_role_types) > 0
            else SampleRoleType.unknown_sample_role.value
        )

    # NOTE: we return a single measurement because we are only have the absorbance data cube measurement at
    # this time, but if there were other measurements to include, we would create multiple measurements here.
    return [
        Measurement(
            measurement_identifier=random_uuid_str(),
            sample_identifier=assert_not_none(injection.get("Label"), "Label"),
            batch_identifier=injection.get("SampleSetID"),
            sample_role_type=sample_role_type,
            written_name=injection.get("SampleName"),
            chromatography_serial_num=injection.get("ColumnSerialNumber")
            or NOT_APPLICABLE,
            autosampler_injection_volume_setting=injection["InjectionVolume"],
            injection_identifier=str(injection["InjectionId"]),
            injection_time=injection["DateAcquired"],
            peaks=[_create_peak(peak) for peak in peaks],
            chromatogram_data_cube=_get_chromatogram(injection),
            device_control_docs=[DeviceControlDoc(device_type=constants.DEVICE_TYPE)],
        )
    ]


def create_measurement_groups(
    injections: list[dict[str, Any]]
) -> list[MeasurementGroup]:
    sample_to_injection: defaultdict[str, list[dict[str, Any]]] = defaultdict(list)
    for injection in injections:
        sample_to_injection[assert_not_none(injection.get("Label"), "Label")].append(
            injection
        )
    return [
        MeasurementGroup(
            measurements=[
                measurement
                for injection in sample_injections
                for measurement in _create_measurements(injection)
            ]
        )
        for sample_injections in sample_to_injection.values()
    ]



from dataclasses import dataclass

from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.revvity_kaleido.kaleido_structure import (
    AnalysisResult,
    BackgroundInfo,
    Data,
    MeasurementInfo,
    Measurements,
    Platemap,
    Results,
)
from allotropy.parsers.utils.values import assert_not_none


@dataclass(frozen=True)
class ResultsV3(Results):
    @classmethod
    def read_barcode(cls, reader: CsvReader) -> str:
        assert_not_none(
            reader.pop_if_match("^Barcode"),
            msg="Unable to find barcode indicator.",
        )

        raw_barcode, *_ = assert_not_none(
            reader.pop_if_match("^.+,"),
            msg="Unable to find barcode value.",
        ).split(",", maxsplit=1)
        return raw_barcode.strip()


def create_data_v3(version: str, reader: CsvReader) -> Data:
    return Data(
        version,
        BackgroundInfo.create(reader),
        ResultsV3.create(reader),
        AnalysisResult.create_results(reader, "Measurement Information"),
        MeasurementInfo.create(
            reader, "Measurement Information", "Plate Type Information"
        ),
        Platemap.create(reader),
        Measurements.create(
            reader, "Details of Measurement Sequence", "Post Processing Sequence"
        ),
    )
